# Copyright 2014 Bevbot LLC
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import copy
import logging
import os
import pwd
import signal
import sys
import subprocess
import time

logger = logging.getLogger(__name__)

POLL_INTERVAL_SECONDS = 1.0


class Runner(object):
    """Runs several commands together as a process group, acting as a watchdog
    while running.

    Limitations:
      * All subprocesses will run in the *same* process group as the parent
        process.
      * All subprocesses will have stdin, stdout, and stderr redirected to
        the OS's /dev/null fd.
      * A Runner cannot be be reused.
    """

    def __init__(self):
        self.commands = {}
        self.watched_procs = {}
        self.running = False
        self.logger = logger

    def is_running(self):
        return self.running

    def add_command(self, command_name, command):
        logger.debug('Adding command "{}": {}'.format(command_name, command))
        if command_name in self.commands:
            raise ValueError('Command already installed')
        self.commands[command_name] = command

    def run(self):
        """Launches all commands, watching their pids."""
        assert not self.is_running(), "Already running!"
        self.running = True

        self.logger.info('Starting commands from pid={}'.format(os.getpid()))
        dev_null_name = getattr(os, 'devnull', '/dev/null')
        dev_null = os.open(dev_null_name, os.O_RDWR)

        # Set sensible env defaults, since supervisor won't.
        path = os.environ.get('PATH', '')
        if sys.argv[0]:
            d = os.path.dirname(sys.argv[0])
            if d not in path.split(':'):
                path = '{}:{}'.format(d, path)

        user = os.environ.get('USER', '')
        if not user:
            try:
                user = pwd.getpwuid(os.getuid()).pw_name
            except KeyError:
                pass

        env = copy.copy(os.environ)
        env['PATH'] = path
        env['USER'] = user

        self.logger.debug('env={}'.format(repr(env)))

        for command_name, command in self.commands.iteritems():
            proc = self._launch_command(command_name, command, dev_null, env)
            self.logger.info('Started {} (pid={})'.format(command_name, proc.pid))
            self.watched_procs[command_name] = proc

        self.watch_commands()

    def watch_commands(self):
        self.logger.info('Watching {} processes.'.format(len(self.commands)))
        while True:
            abort = False
            for command_name, proc in self.watched_procs.iteritems():
                self.logger.debug('Pinging {} (pid={})'.format(command_name, proc.pid))
                proc.poll()
                if proc.returncode is not None:
                    self.logger.info('Process "{}" exited with returncode {}'.format(command_name, proc.returncode))
                    abort = True
            if abort:
                self.abort()
                return
            time.sleep(POLL_INTERVAL_SECONDS)

    def abort(self):
        self.logger.info('Abort called, killing remaining processes ...')
        for command_name, proc in self.watched_procs.iteritems():
            if proc.returncode is None:
                self.logger.info('Killing {} (pid={})'.format(command_name, proc.pid))
                os.killpg(proc.pid, signal.SIGTERM)
        for command_name, proc in self.watched_procs.iteritems():
            self.logger.info('Waiting for {} to exit (pid={}) ...'.format(command_name, proc.pid))
            proc.wait()
            self.logger.info('... done.')
        self.logger.info('All processes exited.')

    def _launch_command(self, command_name, command, dev_null, env=None):
        self.logger.info('Launching command: {}: {}'.format(command_name, command))

        def preexec():
            # Set session id.
            os.setsid()
            # Set umask to default to safe file permissions for root.
            os.umask(0o27)
            # Switch to a "safe" directory.
            os.chdir("/")

        proc = subprocess.Popen(command, stdin=dev_null, stdout=dev_null, stderr=dev_null,
                close_fds=True, shell=True, preexec_fn=preexec,
                env=env)

        return proc

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    r = Runner()
    r.add_command('listen1', 'nc -l 8081')
    r.add_command('listen2', 'nc -l 8082')
    r.run()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Twitter plugin for Kegbot."""

from django.conf import settings
from pykeg.plugin import plugin
from pykeg.plugin import util

from pykeg.core.models import KegbotSite
from pykeg.core.util import SuppressTaskErrors
from socialregistration.contrib.twitter.client import Twitter
import oauth2 as oauth

from . import forms
from . import tasks
from . import views

from unicodedata import normalize

KEY_CONSUMER_KEY = 'consumer_key'
KEY_CONSUMER_SECRET = 'consumer_secret'

KEY_SITE_PROFILE = 'site_profile'
KEY_OAUTH_TOKEN = 'oauth_token'
KEY_OAUTH_TOKEN_SECRET = 'oauth_token_secret'
KEY_TWITTER_NAME = 'twitter_name'
KEY_TWITTER_ID = 'twitter_id'

KEY_SITE_SETTINGS = 'site_settings'

TWITTER_LINK_LEN = 22
TWEET_MAX_LEN = 140
TWEET_MAX_LEN_WITH_URL = TWEET_MAX_LEN - TWITTER_LINK_LEN - 1
TRUNCATE_STR = u'\u2026'


def send_tweet(api, contents):
    """Sends a tweet using `api`."""
    return api.update_status(contents)


def truncate_tweet(tweet, max_len=TWEET_MAX_LEN, truncate_str=TRUNCATE_STR):
    if len(tweet) <= max_len:
        return tweet
    words = tweet.strip().split()
    while True:
        tweet = u' '.join(words)
        candidate = normalize('NFC', tweet + truncate_str)
        if not tweet or len(candidate) <= max_len:
            return candidate
        words = words[:-1]


class TwitterClient(Twitter):
    def set_callback_url(self, url):
        self.callback_url = url

    def get_callback_url(self):
        return self.callback_url


class TwitterPlugin(plugin.Plugin):
    NAME = 'Twitter'
    SHORT_NAME = 'twitter'
    DESCRIPTION = 'Tweet when you pour!'
    URL = 'http://kegbot.org'
    VERSION = '0.0.1-pre'

    def get_admin_settings_view(self):
        return views.admin_settings

    def get_user_settings_view(self):
        return views.user_settings

    def get_extra_admin_views(self):
        return [
            ('redirect/$', 'pykeg.contrib.twitter.views.site_twitter_redirect', 'site_twitter_redirect'),
            ('callback/$', 'pykeg.contrib.twitter.views.site_twitter_callback', 'site_twitter_callback'),
        ]

    def get_extra_user_views(self):
        return [
            ('redirect/$', 'pykeg.contrib.twitter.views.user_twitter_redirect', 'user_twitter_redirect'),
            ('callback/$', 'pykeg.contrib.twitter.views.user_twitter_callback', 'user_twitter_callback'),
        ]

    def get_site_twitter_settings_form(self):
        return self.datastore.load_form(forms.SiteSettingsForm, KEY_SITE_SETTINGS)

    def handle_new_events(self, events):
        for event in events:
            self.handle_event(event)

    def handle_event(self, event):
        self.logger.info('Handling new event: %s' % event.id)
        if util.is_stale(event.time):
            self.logger.info('Event is stale, ignoring: %s' % event.id)
            return

        profile = self.get_site_profile()
        if not profile:
            self.logger.info('No site twitter profile, ignoring.')
            return

        site_settings = self.get_site_twitter_settings_form().initial

        self._issue_system_tweet(event, site_settings, profile)
        if event.user and not event.user.is_guest():
            self._issue_user_tweet(event, site_settings)

    ### Twitter-specific methods

    def get_site_profile(self):
        return self._get_profile(KEY_SITE_PROFILE)

    def get_user_profile(self, user):
        return self._get_profile('user_profile:%s' % user.id)

    def _get_profile(self, datastore_key):
        return self.datastore.get(datastore_key, {})

    def save_site_profile(self, oauth_token, oauth_token_secret,
            twitter_name, twitter_id):
        return self._save_profile(KEY_SITE_PROFILE, oauth_token,
            oauth_token_secret, twitter_name, twitter_id)

    def save_user_profile(self, user, oauth_token, oauth_token_secret,
            twitter_name, twitter_id):
        return self._save_profile('user_profile:%s' % user.id, oauth_token,
            oauth_token_secret, twitter_name, twitter_id)

    def _save_profile(self, datastore_key, oauth_token, oauth_token_secret,
            twitter_name, twitter_id):
        profile = {
            KEY_OAUTH_TOKEN: oauth_token,
            KEY_OAUTH_TOKEN_SECRET: oauth_token_secret,
            KEY_TWITTER_NAME: twitter_name,
            KEY_TWITTER_ID: twitter_id,
        }
        self.datastore.set(datastore_key, profile)
        return profile

    def remove_site_profile(self):
        self.datastore.delete(KEY_SITE_PROFILE)

    def remove_user_profile(self, user):
        self.datastore.delete('user_profile:%s' % user.id)

    def get_credentials(self):
        if settings.EMBEDDED:
            return (
                getattr(settings, 'TWITTER_CONSUMER_KEY', ''),
                getattr(settings, 'TWITTER_CONSUMER_SECRET_KEY', ''),
            )
        return self.datastore.get(KEY_CONSUMER_KEY), self.datastore.get(KEY_CONSUMER_SECRET)

    def set_credentials(self, consumer_key, consumer_secret):
        self.datastore.set(KEY_CONSUMER_KEY, consumer_key)
        self.datastore.set(KEY_CONSUMER_SECRET, consumer_secret)

    def get_user_settings_form(self, user):
        return self.datastore.load_form(forms.UserSettingsForm, 'user_settings:%s' % user.id)

    def get_user_settings(self, user):
        return self.get_user_settings_form(user).initial

    def save_user_settings_form(self, user, form):
        self.datastore.save_form(form, 'user_settings:%s' % user.id)

    def _issue_system_tweet(self, event, settings, site_profile):
        kind = event.kind
        tweet = None
        append_url = bool(settings.get('append_url'))

        if kind == event.DRINK_POURED:
            if not settings.get('tweet_drink_events'):
                self.logger.info('Skipping system tweet for drink event %s: disabled by settings.' % event.id)
                return
            template = settings.get('drink_poured_template')

        elif kind == event.SESSION_STARTED:
            if not settings.get('tweet_session_events'):
                self.logger.info('Skipping system tweet for session start event %s: disabled by settings.' % event.id)
                return
            template = settings.get('session_started_template')

        elif kind == event.SESSION_JOINED:
            if not settings.get('tweet_session_events'):
                self.logger.info('Skipping system tweet for session join event %s: disabled by settings.' % event.id)
                return
            template = settings.get('session_joined_template')

        elif kind == event.KEG_TAPPED:
            if not settings.get('tweet_keg_events'):
                self.logger.info('Skipping system tweet for keg start event %s: disabled by settings.' % event.id)
                return
            template = settings.get('keg_started_template')

        elif kind == event.KEG_ENDED:
            if not settings.get('tweet_keg_events'):
                self.logger.info('Skipping system tweet for keg end event %s: disabled by settings.' % event.id)
                return
            template = settings.get('keg_ended_template')

        else:
            self.logger.info('Skipping unhandled event type: %s' % kind)
            return

        if kind in (event.DRINK_POURED, event.SESSION_JOINED):
            if event.user.is_guest() and not settings.get('include_guests'):
                self.logger.info('Skipping system tweet for event %s: guest pour.' % event.id)
                return

        image_url = None

        if event.drink and event.drink.picture and settings.get('include_pictures'):
            image_url = event.drink.picture.resized.url

        tweet = self._compose_tweet(event, template, append_url)

        if not tweet:
            return

        self.logger.info('Scheduling system tweet: %s' % tweet)
        self._schedule_tweet(tweet, site_profile, image_url)

    def _issue_user_tweet(self, event, site_settings):
        kind = event.kind
        user = event.user
        tweet = None
        append_url = bool(site_settings.get('append_url'))

        if kind not in (event.SESSION_JOINED, event.DRINK_POURED):
            self.logger.info('Skipping user tweet for event %s: unknown event.' % event)
            return

        profile = self.get_user_profile(user)
        if not profile:
            self.logger.info('Skipping user tweet for event %s: no twitter profile.' % event.id)
            return

        user_settings = self.get_user_settings(user)

        if kind == event.DRINK_POURED:
            if not user_settings.get('tweet_drink_events'):
                self.logger.info('Skipping drink tweet for event %s: disabled by user.' % event.id)
                return
            template = site_settings.get('user_drink_poured_template')

        elif kind == event.SESSION_JOINED:
            if not user_settings.get('tweet_session_events'):
                self.logger.info('Skipping session tweet for event %s: disabled by user.' % event.id)
                return
            template = site_settings.get('session_started_template')

        if event.drink.shout:
            template = event.drink.shout

        image_url = None
        if event.drink and event.drink.picture and user_settings.get('include_pictures'):
            image_url = event.drink.picture.resized.url

        tweet = self._compose_tweet(event, template, append_url)

        if not tweet:
            return

        self.logger.info('Scheduling tweet: %s' % tweet)
        self._schedule_tweet(tweet, profile, image_url)

    def _compose_tweet(self, event, template, append_url):
        kbvars = self.get_vars(event)
        tweet = self.expand_template(template, kbvars)
        if append_url:
            url = kbvars['url']
            tweet = truncate_tweet(tweet, TWEET_MAX_LEN_WITH_URL)
            tweet = '%s %s' % (tweet, url)
        return truncate_tweet(tweet)

    def _schedule_tweet(self, tweet, profile, image_url=None):
        if not profile:
            self.logger.warning('Empty profile, skipping tweet.')
            return
        self.logger.info('Scheduling tweet on @%s: %s' % (profile.get('twitter_name'), repr(tweet)))
        consumer_key, consumer_secret = self.get_credentials()
        oauth_token = profile.get(KEY_OAUTH_TOKEN)
        oauth_token_secret = profile.get(KEY_OAUTH_TOKEN_SECRET)

        with SuppressTaskErrors(self.logger):
            tasks.send_tweet.delay(consumer_key, consumer_secret, oauth_token, oauth_token_secret,
                tweet, image_url)

    def get_vars(self, event):
        kbsite = KegbotSite.get()

        username = kbsite.guest_name
        if event.user and not event.user.is_guest():
            username = event.user.username

        url = ''
        if event.drink:
            url = event.drink.short_url()
        elif event.session:
            url = event.session.short_url()

        beverage_name = ''
        if event.drink and (event.drink.keg and event.drink.keg.type):
            beverage_name = event.drink.keg.type.name

        # Must be a keg event if we still have no beer name.
        if event.keg and event.keg.type and not beverage_name:
            beverage_name = event.keg.type.name
            if not url:
                url = event.keg.full_url()

        volume_str = ''
        if event.drink:
            volume_str = kbsite.format_volume(event.drink.volume_ml)

        kbvars = {
            'site_name': kbsite.title,
            'username': username,
            'url': url,
            'volume_str': volume_str,
            'beverage_name': beverage_name,
        }
        return kbvars

    def expand_template(self, template, kbvars):
        """Compiles a template from `kbvars` variables."""
        if not template:
            return ''
        template = template.replace('$DRINKER', kbvars.get('username', ''))
        template = template.replace('$VOLUME', kbvars.get('volume_str', ''))
        template = template.replace('$BEVERAGE', kbvars.get('beverage_name', ''))
        template = template.replace('$BEER', kbvars.get('beverage_name', ''))  # compatibility
        template = template.replace('$SITENAME', kbvars.get('site_name', ''))
        template = template.replace('$URL', kbvars.get('url', ''))
        return template

    def get_client(self):
        consumer_key, consumer_secret = self.get_credentials()
        client = TwitterClient()
        client.consumer = oauth.Consumer(consumer_key, consumer_secret)
        return client



# -*- coding: latin-1 -*-
# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import datetime
import logging
import os
import pytz
import random
import re
import urlparse
from uuid import uuid4
from distutils.version import StrictVersion

from django.contrib.auth.models import AbstractBaseUser
from django.contrib.auth.models import UserManager
from django.core.urlresolvers import reverse
from django.core import validators
from django.core.cache import cache
from django.core.files.storage import default_storage
from django.core.mail import send_mail
from django.db import models
from django.db.models.signals import post_save
from django.db.models.signals import pre_save
from django.utils import timezone

from imagekit.models import ImageSpecField
from imagekit.processors import Adjust
from imagekit.processors import resize

from pykeg.backend import get_kegbot_backend
from pykeg.core import kb_common
from pykeg.core import keg_sizes
from pykeg.core import fields
from pykeg.core import managers
from pykeg.core.util import get_version
from pykeg.util.email import build_message

from kegbot.util import kbjson
from kegbot.util import units
from kegbot.util import util

from pykeg.core.jsonfield import JSONField
from django.utils.translation import ugettext_lazy as _

"""Django models definition for the kegbot database."""

TIMEZONE_CHOICES = ((z, z) for z in pytz.common_timezones)

logger = logging.getLogger(__name__)


class User(AbstractBaseUser):
    """A customized User model based on auth.User.

    Differs from (and does not inherit) AbstractUser because this model:
        - Does not use groups/permissions.
        - Drops `first_name` and `last_name`.
    """

    ### Django AbstractUser fields.

    is_superuser = models.BooleanField(_('superuser status'), default=False,
        help_text=_('Designates that this user has all permissions without '
                    'explicitly assigning them.'))

    username = models.CharField(_('username'), max_length=30, unique=True,
        help_text=_('Required. 30 characters or fewer. Letters, numbers and '
                    '@/./+/-/_ characters'),
        validators=[
            validators.RegexValidator(re.compile('^[\w.@+-]+$'), _('Enter a valid username.'), 'invalid')
    ])
    display_name = models.CharField(default='', max_length=127,
        help_text='Full name, will be shown in some places instead of username')

    email = models.EmailField(_('email address'), blank=True)
    is_staff = models.BooleanField(_('staff status'), default=False,
        help_text=_('Designates whether the user can log into this admin '
                    'site.'))
    is_active = models.BooleanField(_('active'), default=True,
        help_text=_('Designates whether this user should be treated as '
                    'active. Unselect this instead of deleting accounts.'))
    date_joined = models.DateTimeField(_('date joined'), default=timezone.now)

    ### Kegbot fields.

    mugshot = models.ForeignKey('Picture', blank=True, null=True,
        related_name='user_mugshot',
        on_delete=models.SET_NULL)
    activation_key = models.CharField(max_length=128, blank=True,
        null=True,
        help_text='Unguessable token, used to finish registration.')

    objects = UserManager()

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['email']

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')

    def __unicode__(self):
        return self.username

    ### Django-required methods.

    def get_full_name(self):
        return self.display_name

    def get_short_name(self):
        return self.username

    def has_perm(self, perm, obj=None):
        return self.is_superuser

    def has_module_perms(self, app_label):
        return self.is_superuser

    def email_user(self, subject, message, from_email=None, **kwargs):
        """
        Sends an email to this User.
        """
        send_mail(subject, message, from_email, [self.email], **kwargs)

    ### Other methods.

    def get_absolute_url(self):
        return reverse('kb-drinker', kwargs={'username': self.username})

    def is_guest(self):
        return self.username == 'guest'

    def get_stats(self):
        return Stats.get_latest_for_view(user=self)

    def get_api_key(self):
        api_key, new = ApiKey.objects.get_or_create(user=self,
            defaults={'key': ApiKey.generate_key()})
        return api_key.key


def _user_pre_save(sender, instance, **kwargs):
    user = instance
    if not user.display_name:
        user.display_name = user.username

pre_save.connect(_user_pre_save, sender=User)


class Invitation(models.Model):
    """A time-sensitive cookie which can be used to create an account."""
    invite_code = models.CharField(unique=True, max_length=255,
        default=lambda: str(uuid4()),
        help_text='Unguessable token which must be presented to use this invite')
    for_email = models.EmailField(
        help_text='Address this invitation was sent to.')
    invited_date = models.DateTimeField(_('date invited'), auto_now_add=True,
        help_text='Date and time the invitation was sent')
    expires_date = models.DateTimeField(_('date expries'),
        default=lambda: timezone.now() + datetime.timedelta(hours=24),
        help_text='Date and time after which the invitation is considered expired')
    invited_by = models.ForeignKey(User, null=True, on_delete=models.SET_NULL,
        help_text='User that created this invitation, if any.')

    def is_expired(self, now=None):
        if now is None:
            now = timezone.now()
        return now > self.expires_date

    def send(self):
        site = KegbotSite.get()
        base_url = site.reverse_full('registration_register')
        url = base_url + '?invite_code=' + self.invite_code
        context = {
            'site_name': site.title,
            'url': url,
        }
        message = build_message(self.for_email, 'registration/email_invite.html', context)
        if message:
            message.send(fail_silently=True)


class KegbotSite(models.Model):

    VOLUME_DISPLAY_UNITS_CHOICES = (
        ('metric', 'Metric (mL, L)'),
        ('imperial', 'Imperial (oz, pint)'),
    )
    TEMPERATURE_DISPLAY_UNITS_CHOICES = (
        ('f', 'Fahrenheit'),
        ('c', 'Celsius'),
    )
    PRIVACY_CHOICES = (
        ('public', 'Public: Browsing does not require login'),
        ('members', 'Members only: Must log in to browse'),
        ('staff', 'Staff only: Only logged-in staff accounts may browse'),
    )
    REGISTRATION_MODE_CHOICES = (
        ('public', 'Public: Anyone can register.'),
        ('member-invite-only', 'Member Invite: Must be invited by an existing member.'),
        ('staff-invite-only', 'Staff Invite Only: Must be invited by a staff member.'),
    )
    DEFAULT_PRIVACY = 'public'
    DEFAULT_REGISTRATION_MODE = 'public'

    name = models.CharField(max_length=64, unique=True, default='default',
        editable=False)
    server_version = models.CharField(max_length=64, null=True, editable=False)
    is_setup = models.BooleanField(default=False,
        help_text='True if the site has completed setup.',
        editable=False)
    registration_id = models.TextField(max_length=128, editable=False,
        blank=True, default='',
        help_text='A unique id for this system.')

    volume_display_units = models.CharField(max_length=64,
        choices=VOLUME_DISPLAY_UNITS_CHOICES, default='imperial',
        help_text='Unit system to use when displaying volumetric data.')
    temperature_display_units = models.CharField(max_length=64,
        choices=TEMPERATURE_DISPLAY_UNITS_CHOICES, default='f',
        help_text='Unit system to use when displaying temperature data.')
    title = models.CharField(max_length=64, default='My Kegbot',
        help_text='The title of this site.')
    background_image = models.ForeignKey('Picture', blank=True, null=True,
        on_delete=models.SET_NULL,
        help_text='Background for this site.')
    google_analytics_id = models.CharField(blank=True, null=True, max_length=64,
        help_text='Set to your Google Analytics ID to enable tracking. '
        'Example: UA-XXXX-y')
    session_timeout_minutes = models.PositiveIntegerField(
        default=kb_common.DRINK_SESSION_TIME_MINUTES,
        help_text='Maximum time, in minutes, that a session may be idle (no pours) '
        'before it is considered to be finished.  '
        'Recommended value is %s.' % kb_common.DRINK_SESSION_TIME_MINUTES)
    privacy = models.CharField(max_length=63, choices=PRIVACY_CHOICES,
        default=DEFAULT_PRIVACY,
        help_text='Who can view Kegbot data?')
    registration_mode = models.CharField(max_length=63, choices=REGISTRATION_MODE_CHOICES,
        default=DEFAULT_REGISTRATION_MODE,
        help_text='Who can join this Kegbot from the web site?')
    timezone = models.CharField(max_length=255, choices=TIMEZONE_CHOICES,
        default='UTC',
        help_text='Time zone for this system.')

    check_for_updates = models.BooleanField(default=True,
        help_text='Periodically check for updates '
        '(<a href="https://kegbot.org/about/checkin">more info</a>)')

    def __unicode__(self):
        return self.name

    @classmethod
    def get(cls):
        """Gets the default site settings."""
        return KegbotSite.objects.get_or_create(name='default',
            defaults={'is_setup': False, 'server_version': get_version()})[0]

    @classmethod
    def get_installed_version(cls):
        """Returns currently-installed version, or None if not installed."""
        rows = cls.objects.filter(name='default').values('server_version')
        if not rows:
            return None
        return StrictVersion(rows[0].get('server_version', '0.0.0'))

    def get_stats(self):
        return Stats.get_latest_for_view()

    def get_session_timeout_timedelta(self):
        return datetime.timedelta(minutes=self.session_timeout_minutes)

    def base_url(self):
        """Returns the base address of this system."""
        return get_kegbot_backend().get_base_url()

    def full_url(self, path):
        """Returns an absolute URL to the specified path."""
        return urlparse.urljoin(self.base_url(), path)

    def reverse_full(self, *args, **kwargs):
        """Returns an absolute URL to the path reversed by parameters."""
        return self.full_url(reverse(*args, **kwargs))

    def format_volume(self, volume_ml):
        if self.volume_display_units == 'metric':
            if volume_ml < 500:
                return '%d mL' % int(volume_ml)
            return '%.1f L' % (volume_ml / 1000.0)
        else:
            return '%1.f oz' % units.Quantity(volume_ml).InOunces()

    def can_invite(self, user):
        if not user:
            return False
        if self.registration_mode == 'public':
            return True
        if self.registration_mode == 'member-invite-only':
            return not user.is_anonymous()
        if self.registration_mode == 'staff-invite-only':
            return user.is_staff
        return False


class Device(models.Model):
    name = models.CharField(max_length=255, default='Unknown Device')
    created_time = models.DateTimeField(default=timezone.now,
        help_text='Time the device was created.')


class ApiKey(models.Model):
    """Grants access to certain API endpoints to a user via a secret key."""
    user = models.ForeignKey(User, blank=True, null=True,
        help_text='User receiving API access.')
    device = models.ForeignKey(Device, null=True,
        help_text='Device this key is associated with.')
    key = models.CharField(max_length=127, editable=False, unique=True,
        default=lambda: ApiKey.generate_key(),
        help_text='The secret key.')
    active = models.BooleanField(default=True,
        help_text='Whether access by this key is currently allowed.')
    description = models.TextField(blank=True, null=True,
        help_text='Information about this key.')
    created_time = models.DateTimeField(default=timezone.now,
        help_text='Time the key was created.')

    def is_active(self):
        """Returns true if both the key and the key's user are active."""
        return self.active and (not self.user or self.user.is_active)

    def regenerate(self):
        """Discards and regenerates the key."""
        self.key = self.generate_key()
        self.save()

    @classmethod
    def generate_key(cls):
        """Returns a new random key."""
        return '%032x' % random.randint(0, 2 ** 128 - 1)


def _sitesettings_post_save(sender, instance, **kwargs):
    # Privacy settings may have changed.
    cache.clear()
post_save.connect(_sitesettings_post_save, sender=KegbotSite)


class BeverageProducer(models.Model):
    """Information about a beverage producer (brewer, vineyard, etc)."""
    name = models.CharField(max_length=255,
        help_text='Name of the brewer')
    country = fields.CountryField(default='USA',
        help_text='Country of origin')
    origin_state = models.CharField(max_length=128,
        default='', blank=True, null=True,
        help_text='State of origin, if applicable')
    origin_city = models.CharField(max_length=128, default='', blank=True,
        null=True,
        help_text='City of origin, if known')
    is_homebrew = models.BooleanField(default=False)
    url = models.URLField(default='', blank=True, null=True,
        help_text='Brewer\'s home page')
    description = models.TextField(default='', blank=True, null=True,
        help_text='A short description of the brewer')
    picture = models.ForeignKey('Picture', blank=True, null=True,
        on_delete=models.SET_NULL)

    beverage_backend = models.CharField(max_length=255, blank=True, null=True,
        help_text='Future use.')
    beverage_backend_id = models.CharField(max_length=255, blank=True, null=True,
        help_text='Future use.')

    class Meta:
        ordering = ('name',)

    def __unicode__(self):
        return self.name


class Beverage(models.Model):
    """Information about a beverage being served (beer, typically)."""
    TYPE_BEER = 'beer'
    TYPE_WINE = 'wine'
    TYPE_SODA = 'soda'
    TYPE_KOMBUCHA = 'kombucha'
    TYPE_OTHER = 'other'

    TYPES = (
        (TYPE_BEER, 'Beer'),
        (TYPE_WINE, 'Wine'),
        (TYPE_SODA, 'Soda'),
        (TYPE_KOMBUCHA, 'Kombucha'),
        (TYPE_OTHER, 'Other/Unknown'),
    )

    name = models.CharField(max_length=255,
        help_text='Name of the beverage, such as "Potrero Pale".')
    producer = models.ForeignKey(BeverageProducer)

    beverage_type = models.CharField(max_length=32,
        choices=TYPES,
        default=TYPE_BEER)
    style = models.CharField(max_length=255, blank=True, null=True,
        help_text='Beverage style within type, eg "Pale Ale", "Pinot Noir".')
    description = models.TextField(blank=True, null=True,
        help_text='Free-form description of the beverage.')

    picture = models.ForeignKey('Picture', blank=True, null=True,
        help_text='Label image.')
    vintage_year = models.DateField(blank=True, null=True,
        help_text='Date of production, for wines or special/seasonal editions')

    abv_percent = models.FloatField(blank=True, null=True,
        help_text='Alcohol by volume, as percentage (0.0-100.0).')
    calories_per_ml = models.FloatField(blank=True, null=True,
        help_text='Calories per mL of beverage.')
    carbs_per_ml = models.FloatField(blank=True, null=True,
        help_text='Carbohydrates per mL of beverage.')

    original_gravity = models.FloatField(blank=True, null=True,
        help_text='Original gravity (beer only).')
    specific_gravity = models.FloatField(blank=True, null=True,
        help_text='Specific/final gravity (beer only).')
    untappd_beer_id = models.IntegerField(blank=True, null=True,
        help_text='Untappd.com resource ID (beer only).')

    beverage_backend = models.CharField(max_length=255, blank=True, null=True,
        help_text='Future use.')
    beverage_backend_id = models.CharField(max_length=255, blank=True, null=True,
        help_text='Future use.')

    class Meta:
        ordering = ('name',)

    def __unicode__(self):
        return u'{} by {}'.format(self.name, self.producer)


class KegTap(models.Model):
    """A physical tap of beer."""
    class Meta:
        ordering = ('sort_order', 'id')
    name = models.CharField(max_length=128,
        help_text='The display name for this tap, for example, "Main Tap".')
    description = models.TextField(blank=True, null=True,
        help_text='User-visible description for this tap.')
    current_keg = models.OneToOneField('Keg', blank=True, null=True,
        on_delete=models.SET_NULL,
        related_name='current_tap',
        help_text='Keg currently connected to this tap.')
    temperature_sensor = models.ForeignKey('ThermoSensor', blank=True, null=True,
        on_delete=models.SET_NULL,
        help_text='Optional sensor monitoring the temperature at this tap.')
    sort_order = models.PositiveIntegerField(default=0,
        help_text='Position relative to other taps when sorting (0=first).')

    def __unicode__(self):
        return u'{}: {}'.format(self.name, self.current_keg)

    def is_active(self):
        """Returns True if the tap has an active Keg."""
        return self.current_keg is not None

    def current_meter(self):
        """Returns the currently-assigned FlowMeter, or None."""
        try:
            return self.meter
        except FlowMeter.DoesNotExist:
            return None

    def current_toggle(self):
        """Returns the currently-assigned FlowToggle, or None."""
        try:
            return self.toggle
        except FlowToggle.DoesNotExist:
            return None

    def Temperature(self):
        if self.temperature_sensor:
            last_rec = self.temperature_sensor.thermolog_set.all().order_by('-time')
            if last_rec:
                return last_rec[0]
        return None

    @classmethod
    def get_from_meter_name(cls, meter_name):
        try:
            meter = FlowMeter.get_from_meter_name(meter_name)
        except FlowMeter.DoesNotExist, e:
            raise cls.DoesNotExist(e)
        tap = meter.tap
        if not tap:
            raise cls.DoesNotExist('Meter is inactive')
        return tap


class Controller(models.Model):
    name = models.CharField(max_length=128, unique=True,
        help_text='Identifying name for this device; must be unique.')
    model_name = models.CharField(max_length=128, blank=True, null=True,
        help_text='Type of controller (optional).')
    serial_number = models.CharField(max_length=128, blank=True, null=True,
        help_text='Serial number (optional).')

    def __unicode__(self):
        return u'Controller: {}'.format(self.name)


class FlowMeter(models.Model):
    class Meta:
        unique_together = ('controller', 'port_name')
    controller = models.ForeignKey(Controller, related_name='meters',
        help_text='Controller that owns this meter.')
    port_name = models.CharField(max_length=128,
        help_text='Controller-specific data port name for this meter.')
    tap = models.OneToOneField(KegTap, blank=True, null=True,
        related_name='meter',
        help_text='Tap to which this meter is currently bound.')
    ticks_per_ml = models.FloatField(default=kb_common.DEFAULT_TICKS_PER_ML,
        help_text='Flow meter pulses per mL of fluid.  Common values: %s '
        '(SF800), 2.2 (Vision 2000)' % kb_common.DEFAULT_TICKS_PER_ML)

    def meter_name(self):
        return '{}.{}'.format(self.controller.name, self.port_name)

    def __unicode__(self):
        return self.meter_name()

    @classmethod
    def get_or_create_from_meter_name(cls, meter_name):
        try:
            return cls.get_from_meter_name(meter_name)
        except cls.DoesNotExist:
            pass

        idx = meter_name.find('.')
        if idx <= 0:
            raise ValueError('Illegal name')

        controller_name = meter_name[:idx]
        port_name = meter_name[idx + 1:]
        controller = Controller.objects.get_or_create(name=controller_name)[0]
        return cls.objects.get_or_create(controller=controller, port_name=port_name)[0]

    @classmethod
    def get_from_meter_name(cls, meter_name):
        idx = meter_name.find('.')
        if idx <= 0:
            raise cls.DoesNotExist('Illegal meter_name: %s' % repr(meter_name))
        controller_name = meter_name[:idx]
        port_name = meter_name[idx + 1:]

        try:
            controller = Controller.objects.get(name=controller_name)
        except Controller.DoesNotExist:
            raise cls.DoesNotExist('No such controller: %s' % repr(controller_name))

        return cls.objects.get(controller=controller, port_name=port_name)


class FlowToggle(models.Model):
    class Meta:
        unique_together = ('controller', 'port_name')
    controller = models.ForeignKey(Controller, related_name='toggles',
        help_text='Controller that owns this toggle.')
    port_name = models.CharField(max_length=128,
        help_text='Controller-specific data port name for this toggle.')
    tap = models.OneToOneField(KegTap, blank=True, null=True,
        related_name='toggle',
        help_text='Tap to which this toggle is currently bound.')

    def toggle_name(self):
        return u'{}.{}'.format(self.controller.name, self.port_name)

    def __unicode__(self):
        return u'{} (tap: {})'.format(self.toggle_name(), self.tap)

    @classmethod
    def get_or_create_from_toggle_name(cls, toggle_name):
        try:
            return cls.get_from_toggle_name(toggle_name)
        except cls.DoesNotExist:
            pass

        idx = toggle_name.find('.')
        if idx <= 0:
            raise ValueError('Illegal name')

        controller_name = toggle_name[:idx]
        port_name = toggle_name[idx + 1:]
        controller = Controller.objects.get_or_create(name=controller_name)[0]
        return cls.objects.get_or_create(controller=controller, port_name=port_name)[0]

    @classmethod
    def get_from_toggle_name(cls, toggle_name):
        idx = toggle_name.find('.')
        if idx <= 0:
            raise cls.DoesNotExist('Illegal toggle_name: %s' % repr(toggle_name))
        controller_name = toggle_name[:idx]
        port_name = toggle_name[idx + 1:]

        try:
            controller = Controller.objects.get(name=controller_name)
        except Controller.DoesNotExist:
            raise cls.DoesNotExist('No such controller: %s' % repr(controller_name))

        return cls.objects.get(controller=controller, port_name=port_name)


class Keg(models.Model):
    """Record for each physical Keg."""
    type = models.ForeignKey(Beverage,
        on_delete=models.PROTECT,
        help_text='Beverage in this Keg.')
    keg_type = models.CharField(max_length=32,
        choices=keg_sizes.DESCRIPTIONS.items(),
        default=keg_sizes.HALF_BARREL,
        help_text='Keg container type, used to initialize keg\'s full volume')
    served_volume_ml = models.FloatField(default=0, editable=False,
        help_text='Computed served volume.')
    full_volume_ml = models.FloatField(default=0, editable=False,
        help_text='Full volume of this Keg; usually set automatically from keg_type.')
    start_time = models.DateTimeField(default=timezone.now,
        help_text='Time the Keg was first tapped.')
    end_time = models.DateTimeField(default=timezone.now,
        help_text='Time the Keg was finished or disconnected.')
    online = models.BooleanField(default=True, editable=False,
        help_text='True if the keg is currently assigned to a tap.')
    finished = models.BooleanField(default=False, editable=False,
        help_text='True when the Keg has been exhausted or discarded.')
    description = models.CharField(max_length=256, blank=True, null=True,
        help_text='User-visible description of the Keg.')
    spilled_ml = models.FloatField(default=0,
        help_text='Amount of beverage poured without an associated Drink.')
    notes = models.TextField(blank=True, null=True,
        help_text='Private notes about this keg, viewable only by admins.')

    def get_absolute_url(self):
        return reverse('kb-keg', args=(str(self.id),))

    def full_url(self):
        return KegbotSite.get().full_url(self.get_absolute_url())

    def full_volume(self):
        return self.full_volume_ml

    def served_volume(self):
        # Deprecated
        return self.served_volume_ml

    def remaining_volume_ml(self):
        return self.full_volume_ml - self.served_volume_ml - self.spilled_ml

    def percent_full(self):
        result = float(self.remaining_volume_ml()) / float(self.full_volume_ml) * 100
        result = max(min(result, 100), 0)
        return result

    def keg_age(self):
        if self.online:
            end = timezone.now()
        else:
            end = self.end_time
        return end - self.start_time

    def keg_type_description(self):
        return keg_sizes.get_description(self.keg_type)

    def is_empty(self):
        return float(self.remaining_volume_ml()) <= 0

    def previous(self):
        q = Keg.objects.filter(start_time__lt=self.start_time).order_by('-start_time')
        if q.count():
            return q[0]
        return None

    def next(self):
        q = Keg.objects.filter(start_time__gt=self.start_time).order_by('start_time')
        if q.count():
            return q[0]
        return None

    def get_stats(self):
        return Stats.get_latest_for_view(keg=self)

    def Sessions(self):
        sessions_ids = Drink.objects.filter(keg=self.id).values('session').annotate(models.Count('id'))
        pks = [x.get('session') for x in sessions_ids]
        return [DrinkingSession.objects.get(pk=pk) for pk in pks]

    def TopDrinkers(self):
        stats = self.get_stats()
        if not stats:
            return []
        ret = []
        entries = stats.get('volume_by_drinker', {})
        for username, vol in entries.iteritems():
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                continue  # should not happen
            ret.append((vol, user))
        ret.sort(reverse=True)
        return ret

    def __unicode__(self):
        return u'Keg #{} - {}'.format(self.id, self.type)


def _keg_pre_save(sender, instance, **kwargs):
    keg = instance
    # We don't need to do anything if the keg is still online.
    if keg.online:
        return

    # Determine first drink date & set keg start date to it if earlier.
    drinks = keg.drinks.all().order_by('time')
    if drinks:
        drink = drinks[0]
        if drink.time < keg.start_time:
            keg.start_time = drink.time

    # Determine last drink date & set keg end date to it if later.
    drinks = keg.drinks.all().order_by('-time')
    if drinks:
        drink = drinks[0]
        if drink.time > keg.end_time:
            keg.end_time = drink.time

pre_save.connect(_keg_pre_save, sender=Keg)


class Drink(models.Model):
    """ Table of drinks records """
    class Meta:
        get_latest_by = 'time'
        ordering = ('-time',)

    ticks = models.PositiveIntegerField(editable=False,
        help_text='Flow sensor ticks, never changed once recorded.')
    volume_ml = models.FloatField(editable=False,
        help_text='Calculated (or set) Drink volume.')
    time = models.DateTimeField(editable=False,
      help_text='Date and time of pour.')
    duration = models.PositiveIntegerField(blank=True, default=0, editable=False,
        help_text='Time in seconds taken to pour this Drink.')
    user = models.ForeignKey(User, related_name='drinks', editable=False,
        help_text='User responsible for this Drink, or None if anonymous/unknown.')
    keg = models.ForeignKey(Keg, related_name='drinks',
        on_delete=models.PROTECT, editable=False,
        help_text='Keg against which this Drink is accounted.')
    session = models.ForeignKey('DrinkingSession',
        related_name='drinks', null=True, blank=True, editable=False,
        on_delete=models.PROTECT,
        help_text='Session where this Drink is grouped.')
    shout = models.TextField(blank=True, null=True,
        help_text='Comment from the drinker at the time of the pour.')
    tick_time_series = models.TextField(blank=True, null=True, editable=False,
        help_text='Tick update sequence that generated this drink (diagnostic data).')
    picture = models.OneToOneField('Picture', blank=True, null=True,
        on_delete=models.SET_NULL,
        help_text='Picture snapped with this drink.')

    def is_guest_pour(self):
        return self.user is None or self.user.is_guest()

    def get_absolute_url(self):
        return reverse('kb-drink', args=(str(self.id),))

    def short_url(self):
        return KegbotSite.get().reverse_full('kb-drink-short', args=(str(self.id),))

    def Volume(self):
        return units.Quantity(self.volume_ml)

    def calories(self):
        if not self.keg or not self.keg.type:
            return 0
        ounces = self.Volume().InOunces()
        return self.keg.type.calories_oz * ounces

    def __unicode__(self):
        return 'Drink {} by {}'.format(self.id, self.user)


class AuthenticationToken(models.Model):
    """A secret token to authenticate a user, optionally pin-protected."""
    class Meta:
        unique_together = ('auth_device', 'token_value')

    auth_device = models.CharField(max_length=64,
        help_text='Namespace for this token.')
    token_value = models.CharField(max_length=128,
        help_text='Actual value of the token, unique within an auth_device.')
    nice_name = models.CharField(max_length=256, blank=True, null=True,
        help_text='A human-readable alias for the token, for example "Guest Key".')
    pin = models.CharField(max_length=256, blank=True, null=True,
        help_text='A secret value necessary to authenticate with this token.')
    user = models.ForeignKey(User, blank=True, null=True,
        related_name='tokens',
        help_text='User in possession of and authenticated by this token.')
    enabled = models.BooleanField(default=True,
        help_text='Whether this token is considered active.')
    created_time = models.DateTimeField(auto_now_add=True,
        help_text='Date token was first added to the system.')
    expire_time = models.DateTimeField(blank=True, null=True,
        help_text='Date after which token is treated as disabled.')

    def __unicode__(self):
        auth_device = self.auth_device
        if auth_device == 'core.rfid':
            auth_device = 'RFID'
        elif auth_device == 'core.onewire':
            auth_device = 'OneWire'

        ret = u'{} {}'.format(auth_device, self.token_value)
        if self.nice_name:
            ret += u' ({})'.format(self.nice_name)
        return ret

    def get_auth_device(self):
        auth_device = self.auth_device
        if auth_device == 'core.rfid':
            auth_device = 'RFID'
        elif auth_device == 'core.onewire':
            auth_device = 'OneWire'
        return auth_device

    def IsAssigned(self):
        return self.user is not None

    def IsActive(self):
        if not self.enabled:
            return False
        if not self.expire_time:
            return True
        return timezone.now() < self.expire_time


def _auth_token_pre_save(sender, instance, **kwargs):
    if instance.auth_device in kb_common.AUTH_MODULE_NAMES_HEX_VALUES:
        instance.token_value = instance.token_value.lower()

pre_save.connect(_auth_token_pre_save, sender=AuthenticationToken)


class DrinkingSession(models.Model):
    """A collection of contiguous drinks. """
    class Meta:
        get_latest_by = 'start_time'
        ordering = ('-start_time',)
    start_time = models.DateTimeField()
    end_time = models.DateTimeField()
    volume_ml = models.FloatField(default=0)

    objects = managers.SessionManager()
    name = models.CharField(max_length=256, blank=True, null=True)

    def __unicode__(self):
        return 'Session #{}: {}'.format(self.id, self.start_time)

    def Duration(self):
        return self.end_time - self.start_time

    def _AddDrinkNoSave(self, drink):
        session_delta = KegbotSite.get().get_session_timeout_timedelta()
        session_end = drink.time + session_delta

        if self.start_time > drink.time:
            self.start_time = drink.time
        if self.end_time < session_end:
            self.end_time = session_end
        self.volume_ml += drink.volume_ml

    def AddDrink(self, drink):
        self._AddDrinkNoSave(drink)
        self.save()

    def short_url(self):
        return KegbotSite.get().reverse_full('kb-session-short', args=(str(self.id),))

    def full_url(self):
        return KegbotSite.get().full_url(self.get_absolute_url())

    def get_highlighted_picture(self):
        pictures = self.pictures.all().order_by('-time')
        if pictures:
            return pictures[0]
        return None

    def get_non_highlighted_pictures(self):
        pictures = self.pictures.all().order_by('-time')
        if pictures:
            return pictures[1:]
        return []

    def get_absolute_url(self):
        dt = timezone.localtime(self.start_time)
        return reverse('kb-session-detail', args=(), kwargs={
            'year': dt.year,
            'month': dt.month,
            'day': dt.day,
            'pk': self.pk})

    def get_stats(self):
        return Stats.get_latest_for_view(session=self)

    def summarize_drinkers(self):
        stats = self.get_stats()
        volmap = stats.get('volume_by_drinker', {})
        names = [x for x in reversed(sorted(volmap, key=volmap.get))]

        if 'guest' in names:
            guest_trailer = ' (and possibly others)'
        else:
            guest_trailer = ''

        num = len(names)
        if num == 0:
            return 'no known drinkers'
        elif num == 1:
            ret = names[0]
        elif num == 2:
            ret = '{} and {}'.format(*names)
        elif num == 3:
            ret = '{}, {} and {}'.format(*names)
        else:
            if guest_trailer:
                return '%s, %s and at least %i others' % (names[0], names[1], num - 2)
            else:
                return '%s, %s and %i others' % (names[0], names[1], num - 2)

        return '%s%s' % (ret, guest_trailer)

    def GetTitle(self):
        if self.name:
            return self.name
        else:
            if self.id:
                return 'Session %s' % (self.id,)
            else:
                # Not yet saved.
                return 'New Session'

    def IsActiveNow(self):
        return self.IsActive(timezone.now())

    def IsActive(self, now):
        return self.end_time > now

    def Rebuild(self):
        """Recomputes start time, end time, and volume, based on current drinks.

        This method should be called after changing the set of drinks
        belonging to this session.

        This method has no effect on statistics; see stats module.
        """
        self.volume_ml = 0

        drinks = self.drinks.all()
        if not drinks:
            self.delete()
            return

        session_delta = KegbotSite.get().get_session_timeout_timedelta()
        min_time = None
        max_time = None
        for d in drinks:
            self.AddDrink(d)
            if min_time is None or d.time < min_time:
                min_time = d.time
            if max_time is None or d.time > max_time:
                max_time = d.time
        self.start_time = min_time
        self.end_time = max_time + session_delta
        self.save()

    @classmethod
    def AssignSessionForDrink(cls, drink):
        # Return existing session if already assigned.
        if drink.session:
            return drink.session

        # Return last session if one already exists
        q = DrinkingSession.objects.all().order_by('-end_time')[:1]
        if q and q[0].IsActive(drink.time):
            session = q[0]
            session.AddDrink(drink)
            drink.session = session
            drink.save()
            return session

        # Create a new session
        session = cls(start_time=drink.time, end_time=drink.time)
        session.save()
        session.AddDrink(drink)
        drink.session = session
        drink.save()
        return session


class ThermoSensor(models.Model):
    raw_name = models.CharField(max_length=256)
    nice_name = models.CharField(max_length=128)

    def __unicode__(self):
        if self.nice_name:
            return u'{} ({})'.format(self.nice_name, self.raw_name)
        return self.raw_name

    def LastLog(self):
        try:
            return self.thermolog_set.latest()
        except Thermolog.DoesNotExist:
            return None


class Thermolog(models.Model):
    """ A log from an ITemperatureSensor device of periodic measurements. """
    class Meta:
        get_latest_by = 'time'
        ordering = ('-time',)

    sensor = models.ForeignKey(ThermoSensor)
    temp = models.FloatField()
    time = models.DateTimeField()

    def __unicode__(self):
        return u'%.2f C / %.2f F [%s]' % (self.TempC(), self.TempF(), self.time)

    def TempC(self):
        return self.temp

    def TempF(self):
        return util.CtoF(self.temp)


class Stats(models.Model):
    """Derived statistics.

    Multiple "views" of statistics are stored in this table, for example,
    only statistics for a particular user in a particular session.  The view
    is determined by the columns (user, keg, session), any combination of which
    may be null.

    Stats are temporal on a per-drink basis: for every drink that is recorded,
    a new row is generated for each view (2^3 rows total).

    See stats.generate(..) for generation details.
    """
    time = models.DateTimeField(default=timezone.now)
    stats = JSONField(dump_kwargs={'cls': kbjson.JSONEncoder})
    drink = models.ForeignKey(Drink)

    is_first = models.BooleanField(default=False,
        help_text='True if this is the most first record for the view.')

    # Any combination of these fields is allowed.
    user = models.ForeignKey(User, related_name='stats', null=True)
    keg = models.ForeignKey(Keg, related_name='stats', null=True)
    session = models.ForeignKey(DrinkingSession, related_name='stats', null=True)

    class Meta:
        get_latest_by = 'id'
        unique_together = ('drink', 'user', 'keg', 'session')

    @classmethod
    def apply_usernames(cls, stats):
        """Given a stats dictionary, translate numeric user ids to usernames."""
        def safe_get_user(pk):
            try:
                return User.objects.get(pk=pk)
            except (User.DoesNotExist, ValueError):
                return None
        orig = stats.get('registered_drinkers', [])
        if orig:
            stats['registered_drinkers'] = [safe_get_user(pk).username for pk in orig if safe_get_user(pk)]

        orig = stats.get('volume_by_drinker', util.AttrDict())
        if orig:
            stats['volume_by_drinker'] = util.AttrDict(
                (safe_get_user(pk).username, val) for pk, val in orig.iteritems() if safe_get_user(pk))

    @classmethod
    def get_latest_for_view(cls, user=None, keg=None, session=None):
        """Returns the most recent stats data for the (user, keg, session) tuple.

        Returns an empty dict if no stats available for this view.
        """
        try:
            stats = cls.objects.filter(user=user, keg=keg, session=session).order_by('-id')[0].stats
        except IndexError:
            stats = {}
        cls.apply_usernames(stats)
        return util.AttrDict(stats)


class SystemEvent(models.Model):
    class Meta:
        ordering = ('-id',)
        get_latest_by = 'time'

    DRINK_POURED = 'drink_poured'
    SESSION_STARTED = 'session_started'
    SESSION_JOINED = 'session_joined'
    KEG_TAPPED = 'keg_tapped'
    KEG_VOLUME_LOW = 'keg_volume_low'
    KEG_ENDED = 'keg_ended'

    KINDS = (
        (DRINK_POURED, 'Drink poured'),
        (SESSION_STARTED, 'Session started'),
        (SESSION_JOINED, 'User joined session'),
        (KEG_TAPPED, 'Keg tapped'),
        (KEG_VOLUME_LOW, 'Keg volume low'),
        (KEG_ENDED, 'Keg ended'),
    )

    kind = models.CharField(max_length=255, choices=KINDS,
        help_text='Type of event.')
    time = models.DateTimeField(help_text='Time of the event.')
    user = models.ForeignKey(User, blank=True, null=True,
        related_name='events',
        help_text='User responsible for the event, if any.')
    drink = models.ForeignKey(Drink, blank=True, null=True,
        related_name='events',
        help_text='Drink involved in the event, if any.')
    keg = models.ForeignKey(Keg, blank=True, null=True,
        related_name='events',
        help_text='Keg involved in the event, if any.')
    session = models.ForeignKey(DrinkingSession, blank=True, null=True,
        related_name='events',
        help_text='Session involved in the event, if any.')

    objects = managers.SystemEventManager()

    def __unicode__(self):
        if self.kind == self.DRINK_POURED:
            ret = u'Drink {} poured'.format(self.drink.id)
        elif self.kind == self.SESSION_STARTED:
            ret = u'Session {} started by drink {}'.format(self.session.id, self.drink.id)
        elif self.kind == self.SESSION_JOINED:
            ret = u'Session {} joined by {} (drink {})'.format(self.session.id,
                self.user.username, self.drink.id)
        elif self.kind == self.KEG_TAPPED:
            ret = u'Keg {} tapped'.format(self.keg.id)
        elif self.kind == self.KEG_VOLUME_LOW:
            ret = u'Keg {} volume low'.format(self.keg.id)
        elif self.kind == self.KEG_ENDED:
            ret = u'Keg {} ended'.format(self.keg.id)
        else:
            ret = u'Unknown event type ({})'.format(self.kind)
        return u'Event {}: {}'.format(self.id, ret)

    @classmethod
    def build_events_for_keg(cls, keg):
        '''Generates and returns system events for the keg.'''
        events = []
        if keg.online:
            q = keg.events.filter(kind=cls.KEG_TAPPED)
            if q.count() == 0:
                e = keg.events.create(kind=cls.KEG_TAPPED, time=keg.start_time, keg=keg)
                e.save()
                events.append(e)

        if not keg.online:
            q = keg.events.filter(kind=cls.KEG_ENDED)
            if q.count() == 0:
                e = keg.events.create(kind=cls.KEG_ENDED, time=keg.end_time, keg=keg)
                e.save()
                events.append(e)

        return events

    @classmethod
    def build_events_for_drink(cls, drink):
        keg = drink.keg
        session = drink.session
        user = drink.user

        events = cls.build_events_for_keg(keg)

        if session:
            q = session.events.filter(kind=cls.SESSION_STARTED)
            if q.count() == 0:
                e = session.events.create(kind=cls.SESSION_STARTED,
                    time=session.start_time, drink=drink, user=user)
                e.save()
                events.append(e)

        if user:
            q = user.events.filter(kind=cls.SESSION_JOINED, session=session)
            if q.count() == 0:
                e = user.events.create(kind=cls.SESSION_JOINED,
                    time=drink.time, session=session, drink=drink, user=user)
                e.save()
                events.append(e)

        e = drink.events.create(kind=cls.DRINK_POURED,
            time=drink.time, drink=drink, user=user, keg=keg,
            session=session)
        e.save()
        events.append(e)

        volume_now = keg.remaining_volume_ml()
        volume_before = volume_now + drink.volume_ml
        threshold = keg.full_volume_ml * kb_common.KEG_VOLUME_LOW_PERCENT

        if volume_now <= threshold and volume_before > threshold:
            e = drink.events.create(kind=cls.KEG_VOLUME_LOW,
                time=drink.time, drink=drink, user=user, keg=keg,
                session=session)
            e.save()
            events.append(e)

        return events


def _pics_file_name(instance, filename, now=None, uuid_str=None):
    if not now:
        now = timezone.now()
    if not uuid_str:
        uuid_str = str(uuid4()).replace('-', '')
    date_str = now.strftime('%Y%m%d%H%M%S')
    ext = os.path.splitext(filename)[1]
    new_filename = '%s-%s%s' % (date_str, uuid_str, ext)

    return os.path.join('pics', new_filename)


class Picture(models.Model):
    image = models.ImageField(upload_to=_pics_file_name,
        help_text='The image')
    resized = ImageSpecField(source='image',
        processors=[resize.ResizeToFit(1024, 1024)],
        format='JPEG',
        options={'quality': 100})
    resized_png = ImageSpecField(source='image',
        processors=[resize.ResizeToFit(1024, 1024)],
        format='PNG',
        options={'quality': 100})
    thumbnail = ImageSpecField(source='image',
        processors=[Adjust(contrast=1.2, sharpness=1.1), resize.SmartResize(128, 128)],
        format='JPEG',
        options={'quality': 90})
    thumbnail_png = ImageSpecField(source='image',
        processors=[Adjust(contrast=1.2, sharpness=1.1), resize.SmartResize(128, 128)],
        format='PNG',
        options={'quality': 90})

    time = models.DateTimeField(default=timezone.now,
        help_text='Time/date of image capture')
    caption = models.TextField(blank=True, null=True,
        help_text='Caption for the picture, if any.')
    user = models.ForeignKey(User, blank=True, null=True,
        related_name='pictures',
        help_text='User that owns/uploaded this picture')
    keg = models.ForeignKey(Keg, blank=True, null=True,
        related_name='pictures',
        on_delete=models.SET_NULL,
        help_text='Keg this picture was taken with, if any.')
    session = models.ForeignKey(DrinkingSession, blank=True, null=True,
        related_name='pictures',
        on_delete=models.SET_NULL,
        help_text='Session this picture was taken with, if any.')

    def __unicode__(self):
        return u'Picture: {}'.format(self.image)

    def get_caption(self):
        if self.caption:
            return self.caption
        elif self.drink:
            if self.user:
                return u'{} pouring drink {}'.format(self.user.username, self.drink.id)
            else:
                return u'An unknown drinker pouring drink {}'.format(self.drink.id)
        return ''

    def erase_and_delete(self):
        try:
            for spec in ('resized', 'resized_png', 'thumbnail', 'thumbnail_png', 'image'):
                image_file = getattr(self, spec)
                exists = bool(image_file)
                if not exists:
                    logger.debug('erase_and_delete: image.id={} spec={}: does not exist'.format(
                        self.id, spec))
                    continue
                logger.debug('erase_and_delete: image.id={} spec={}: deleting ...'.format(
                    self.id, spec))

                try:
                    try:
                        default_storage.delete(image_file.path)
                    except NotImplementedError:
                        default_storage.delete(image_file.name)
                    logger.debug('erase_and_delete: image.id={} spec={}: deleted'.format(
                        self.id, spec))
                except IOError as e:
                    logger.warning('erase_and_delete: image.id={} spec={}: error: {}'.format(
                        self.id, spec, e))
        finally:
            self.delete()


class NotificationSettings(models.Model):
    """Stores a user's notification settings for a notification backend."""

    class Meta:
        unique_together = ('user', 'backend')

    user = models.ForeignKey(User,
        help_text='User for these settings.')
    backend = models.CharField(max_length=255,
        help_text='Notification backend (dotted path) for these settings.')
    keg_tapped = models.BooleanField(default=True,
        help_text='Sent when a keg is activated.')
    session_started = models.BooleanField(default=False,
        help_text='Sent when a new drinking session starts.')
    keg_volume_low = models.BooleanField(default=False,
        help_text='Sent when a keg becomes low.')
    keg_ended = models.BooleanField(default=False,
        help_text='Sent when a keg has been taken offline.')


class PluginData(models.Model):
    """Key/value JSON data store for plugins."""

    class Meta:
        unique_together = ('plugin_name', 'key')

    plugin_name = models.CharField(max_length=127,
        help_text='Plugin short name')
    key = models.CharField(max_length=127)
    value = JSONField()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from pykeg.core.management.commands.common import RunnerCommand
from optparse import make_option


class Command(RunnerCommand):
    help = u'Runs background task queue workers.'
    pidfile_name = 'kegbot_run_all.pid'

    option_list = RunnerCommand.option_list + (
        make_option('--gunicorn_options', action='store', dest='gunicorn_options', default='-w 3',
            help='Specifies extra options to pass to gunicorn.'),
    )

    def get_commands(self, options):
        ret = []
        logs_dir = options.get('logs_dir')

        workers_command = 'kegbot run_workers'
        if logs_dir:
            workers_command += ' --logs_dir={}'.format(logs_dir)
        ret.append(('workers', workers_command))

        extra_options = options.get('gunicorn_options', '')
        ret.append(('guincorn', 'kegbot run_gunicorn --settings=pykeg.settings ' + extra_options))
        return ret



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.test import TransactionTestCase

from pykeg.plugin import datastore
from pykeg.core import models


class DatastoreTestCase(TransactionTestCase):

    def test_model_datastore(self):
        ds = datastore.ModelDatastore(plugin_name='test')

        self.assertEqual(0, models.PluginData.objects.all().count())

        ds.set('foo', 'bar')
        q = models.PluginData.objects.all()
        self.assertEqual(1, q.count())
        self.assertEqual('test', q[0].plugin_name)
        self.assertEqual('foo', q[0].key)
        self.assertEqual('bar', q[0].value)

        # Setting to 'None' clears value
        ds.set('foo', None)
        self.assertEqual(0, models.PluginData.objects.all().count())

        # Complex types survive.
        ds.set('obj', {'asdf': 123, 'foo': None})
        self.assertEqual({'asdf': 123, 'foo': None}, ds.get('obj'))

    def test_in_memory_datastore(self):
        ds = datastore.InMemoryDatastore(plugin_name='test')

        self.assertEqual(0, len(ds.data))

        ds.set('foo', 'bar')
        self.assertEqual(1, len(ds.data))
        self.assertEqual('bar', ds.data['test:foo'])

        # Setting to 'None' clears value
        ds.set('foo', None)
        self.assertEqual(0, len(ds.data))

        # Complex types survive.
        ds.set('obj', {'asdf': 123, 'foo': None})
        self.assertEqual({'asdf': 123, 'foo': None}, ds.get('obj'))



from django.conf import settings
from django.conf.urls import patterns
from django.conf.urls import url

urlpatterns = patterns('pykeg.web.kegadmin.views',
    ### main page
    url(r'^$', 'dashboard', name='kegadmin-dashboard'),
    url(r'^settings/general/$', 'general_settings', name='kegadmin-main'),
    url(r'^settings/location/$', 'location_settings', name='kegadmin-location-settings'),
    url(r'^settings/advanced/$', 'advanced_settings', name='kegadmin-advanced-settings'),

    url(r'^export/$', 'export', name='kegadmin-export'),

    url(r'^beers/$', 'beverages_list', name='kegadmin-beverages'),
    url(r'^beers/add/$', 'beverage_add', name='kegadmin-add-beverage'),
    url(r'^beers/(?P<beer_id>\d+)/$', 'beverage_detail', name='kegadmin-edit-beverage'),

    url(r'^devices/link/$', 'link_device', name='kegadmin-link-device'),

    url(r'^kegs/$', 'keg_list', name='kegadmin-kegs'),
    url(r'^kegs/add/$', 'keg_add', name='kegadmin-add-keg'),
    url(r'^kegs/(?P<keg_id>\d+)/$', 'keg_detail', name='kegadmin-edit-keg'),

    url(r'^brewers/$', 'beverage_producer_list', name='kegadmin-beverage-producers'),
    url(r'^brewers/add/$', 'beverage_producer_add', name='kegadmin-add-beverage-producer'),
    url(r'^brewers/(?P<brewer_id>\d+)/$', 'beverage_producer_detail', name='kegadmin-edit-beverage-producer'),

    url(r'^controllers/$', 'controller_list', name='kegadmin-controllers'),
    url(r'^controllers/(?P<controller_id>\d+)/$', 'controller_detail', name='kegadmin-edit-controller'),

    url(r'^taps/$', 'tap_list', name='kegadmin-taps'),
    url(r'^taps/create/$', 'add_tap', name='kegadmin-add-tap'),
    url(r'^taps/(?P<tap_id>\d+)/$', 'tap_detail', name='kegadmin-edit-tap'),

    url(r'^users/$', 'user_list', name='kegadmin-users'),
    url(r'^users/(?P<user_id>\d+)/$', 'user_detail', name='kegadmin-edit-user'),

    url(r'^drinks/$', 'drink_list', name='kegadmin-drinks'),
    url(r'^drinks/(?P<drink_id>\d+)/$', 'drink_edit', name='kegadmin-edit-drink'),

    url(r'^tokens/$', 'token_list', name='kegadmin-tokens'),
    url(r'^tokens/create/$', 'add_token', name='kegadmin-add-token'),
    url(r'^tokens/(?P<token_id>\d+)/$', 'token_detail', name='kegadmin-edit-token'),

    url(r'^autocomplete/beverage/$', 'autocomplete_beverage',
      name='kegadmin-autocomplete-beverage'),
    url(r'^autocomplete/user/$', 'autocomplete_user',
      name='kegadmin-autocomplete-user'),
    url(r'^autocomplete/token/$', 'autocomplete_token',
      name='kegadmin-autocomplete-token'),

    url(r'^plugin/(?P<plugin_name>\w+)/$', 'plugin_settings', name='kegadmin-plugin-settings'),
)

if not settings.EMBEDDED:
    urlpatterns += patterns('pykeg.web.kegadmin.views',
        url(r'^email/$', 'email', name='kegadmin-email'),
        url(r'^logs/$', 'logs', name='kegadmin-logs'),
        url(r'^users/create/$', 'add_user', name='kegadmin-add-user'),
        url(r'^workers/$', 'workers', name='kegadmin-workers'),
    )

from pykeg.plugin import util
if util.get_plugins():
    urlpatterns += util.get_admin_urls()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for use in tests."""

import datetime
import os

from pykeg.backend.backends import KegbotBackend

from django.conf import settings
from django.utils import timezone

from django_nose import NoseTestSuiteRunner


TESTDATA_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../testdata/'))


def get_filename(f):
    return os.path.join(TESTDATA_DIR, f)


def make_datetime(*args):
    if settings.USE_TZ:
        return datetime.datetime(*args, tzinfo=timezone.utc)
    else:
        return datetime.datetime(*args)


class TestBackend(KegbotBackend):
    def get_base_url(self):
        static_url = getattr(settings, 'KEGBOT_BASE_URL', None)
        if static_url:
            return static_url.rstrip('/')
        return 'http://localhost:1234'


class KegbotTestSuiteRunner(NoseTestSuiteRunner):
    def __init__(self, *args, **kwargs):
        # Run all celery tasks synchronously.
        settings.CELERY_ALWAYS_EAGER = True
        super(KegbotTestSuiteRunner, self).__init__(*args, **kwargs)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import os
import signal
import sys
from optparse import make_option
from django.core.management.base import BaseCommand
from pykeg.util.runner import Runner


def progbar(title, pos, total, width=40):
    """Prints a progress bar to stdout.

    Args
      title: title to show next to progress bar
      pos: current position (integer)
      total: total positions (integer)
      width: width of the progres bar, in characters
    """
    if not total:
        chars = width
    else:
        chars = int((float(pos) / total) * width)
    rem = width - chars
    inner = '+' * chars + ' ' * rem
    sys.stdout.write('%-30s  [%s] %i/%i\r' % (title, inner, pos, total))
    sys.stdout.flush()


def check_and_create_pid_file(pid_file):
    if os.path.exists(pid_file):
        print 'Error: already running ({})'.format(pid_file)
        sys.exit(1)

    f = open(pid_file, 'w')
    f.write('{}\n'.format(os.getpid()))
    f.close()


class check_pidfile(object):
    """Context manager that creates a pidfile, or fails if it exists."""

    def __init__(self, pid_file):
        self.pid_file = pid_file

    def __enter__(self):
        if os.path.exists(self.pid_file):
            raise RuntimeError('Error: already running ({})'.format(self.pid_file))
        f = open(self.pid_file, 'w')
        f.write('{}\n'.format(os.getpid()))
        f.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.unlink(self.pid_file)


class RunnerCommand(BaseCommand):
    """Command that runs several subcommands as a watched group."""
    requires_model_validation = False

    option_list = BaseCommand.option_list + (
        make_option('--logs_dir', action='store', dest='logs_dir', default='',
            help='Specifies the directory for log files.  If empty, logging disabled.'),
        make_option('--pidfile_dir', action='store', dest='pidfile_dir', default='/tmp',
            help='PID file for this program.'),
    )

    pidfile_name = None

    def get_commands(self, options):
        """Returns iterable of (command_name, command)."""
        raise NotImplementedError

    def handle(self, *args, **options):
        pidfile = os.path.join(options['pidfile_dir'], self.pidfile_name)
        with check_pidfile(pidfile):
            r = Runner()

            def exit_fn(signum, frame):
                r.abort()
                sys.exit(1)

            signal.signal(signal.SIGTERM, exit_fn)
            signal.signal(signal.SIGINT, exit_fn)

            for command_name, command in self.get_commands(options):
                r.add_command(command_name, command)

            r.run()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb RESTful API views."""

import datetime
import logging
from functools import wraps

from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.contrib.auth.forms import AuthenticationForm
from django.utils import timezone

from django.http import Http404
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

from kegbot.api import kbapi

from pykeg import backend
from pykeg.core import keg_sizes
from pykeg.core import models
from pykeg.core import util as core_util
from pykeg.proto import protolib
from pykeg.web.api import devicelink
from pykeg.web.api import forms
from pykeg.web.api import util
from pykeg.web.kegadmin.forms import ChangeKegForm
from pykeg.web.kegadmin.forms import ControllerForm
from pykeg.web.kegadmin.forms import FlowToggleForm
from pykeg.web.kegadmin.forms import NewFlowMeterForm
from pykeg.web.kegadmin.forms import UpdateFlowMeterForm

_LOGGER = logging.getLogger(__name__)

RESULT_OK = {'result': 'ok'}

### Decorators


def auth_required(view_func):
    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)
    util.set_needs_auth(wrapped_view)
    return wraps(view_func)(wrapped_view)

### Helpers


def _form_errors(form):
    ret = {}
    for field in form:
        if field.errors:
            name = field.html_name
            ret[name] = []
            for error in field.errors:
                ret[name].append(error)
    return ret

### Endpoints


def all_kegs(request):
    return models.Keg.objects.all().order_by('-start_time')


def all_drinks(request, limit=100):
    qs = models.Drink.objects.all()
    if 'start' in request.GET:
        try:
            start = int(request.GET['start'])
            qs = qs.filter(id__lte=start)
        except ValueError:
            pass
    qs = qs.order_by('-id')
    qs = qs[:limit]
    return qs


def get_drink(request, drink_id):
    drink = get_object_or_404(models.Drink, id=drink_id)
    return protolib.ToProto(drink, full=True)


@csrf_exempt
@auth_required
def all_controllers(request):
    if request.method == 'POST':
        form = ControllerForm(request.POST)
        if form.is_valid():
            return form.save()
        else:
            errors = _form_errors(form)
            raise kbapi.BadRequestError(errors)

    return models.Controller.objects.all()


@csrf_exempt
@auth_required
def get_controller(request, controller_id):
    controller = get_object_or_404(models.Controller, id=controller_id)

    if request.method == 'DELETE':
        controller.delete()
        return RESULT_OK

    elif request.method == 'POST':
        form = ControllerForm(request.POST, instance=controller)
        if form.is_valid():
            controller = form.save()
        else:
            errors = _form_errors(form)
            raise kbapi.BadRequestError(errors)

    return protolib.ToProto(controller, full=True)


@csrf_exempt
@auth_required
def all_flow_meters(request):
    if request.method == 'POST':
        form = NewFlowMeterForm(request.POST)
        if form.is_valid():
            return form.save()
        else:
            errors = _form_errors(form)
            raise kbapi.BadRequestError(errors)

    return models.FlowMeter.objects.all()


@csrf_exempt
@auth_required
def get_flow_meter(request, flow_meter_id):
    meter = get_object_or_404(models.FlowMeter, id=flow_meter_id)

    if request.method == 'DELETE':
        meter.delete()
        return RESULT_OK

    elif request.method == 'POST':
        form = UpdateFlowMeterForm(request.POST, instance=meter)
        if form.is_valid():
            meter = form.save()
        else:
            errors = _form_errors(form)
            raise kbapi.BadRequestError(errors)

    return protolib.ToProto(meter, full=True)


@csrf_exempt
@auth_required
def all_flow_toggles(request):
    if request.method == 'POST':
        form = FlowToggleForm(request.POST)
        if form.is_valid():
            return form.save()
        else:
            errors = _form_errors(form)
            raise kbapi.BadRequestError(errors)

    return models.FlowToggle.objects.all()


@csrf_exempt
@auth_required
def get_flow_toggle(request, flow_toggle_id):
    toggle = get_object_or_404(models.FlowToggle, id=flow_toggle_id)

    if request.method == 'DELETE':
        toggle.delete()
        return RESULT_OK

    elif request.method == 'POST':
        form = FlowToggleForm(request.POST, instance=toggle)
        if form.is_valid():
            toggle = form.save()
        else:
            errors = _form_errors(form)
            raise kbapi.BadRequestError(errors)

    return protolib.ToProto(toggle, full=True)


@csrf_exempt
@auth_required
def pictures(request):
    if request.method != 'POST':
        raise Http404('Method not supported')
    pic = models.Picture.objects.create(
        image=request.FILES['photo'],
    )
    return protolib.ToProto(pic, full=True)


@csrf_exempt
@auth_required
def add_drink_photo(request, drink_id):
    if request.method != 'POST':
        raise Http404('Method not supported')
    drink = get_object_or_404(models.Drink, id=drink_id)
    pic = _save_pour_pic(request, drink)
    return protolib.ToProto(pic, full=True)


def _save_pour_pic(request, drink):
    pic = models.Picture.objects.create(
        image=request.FILES['photo'],
        user=drink.user,
        keg=drink.keg,
        session=drink.session
    )
    # TODO(mikey): Should we do anything with a previously-saved
    # picture here?
    drink.picture = pic
    drink.save()
    return pic


def get_session(request, session_id):
    session = get_object_or_404(models.DrinkingSession, id=session_id)
    return protolib.ToProto(session, full=True)


def get_session_stats(request, session_id):
    session = get_object_or_404(models.DrinkingSession, id=session_id)
    return session.get_stats()


@auth_required
def get_status(request):
    try:
        session = current_session(request)
    except Http404:
        session = None

    controllers = models.Controller.objects.all()
    drinks = models.Drink.objects.all()[:5]
    events = models.SystemEvent.objects.all()[:5]
    kegs = models.Keg.objects.all().filter(online=True)
    meters = models.FlowMeter.objects.all()
    sound_events = []   # deprecated
    taps = models.KegTap.objects.all()
    toggles = models.FlowToggle.objects.all()

    current_users = set()
    if session:
        for stat in models.Stats.objects.filter(session=session):
            user = stat.user
            if not user.is_guest():
                current_users.add(user)

    title = models.KegbotSite.get().title
    version = core_util.get_version()

    response = protolib.GetSyncResponse(
        active_kegs=kegs, active_session=session, active_users=current_users,
        controllers=controllers, drinks=drinks, events=events, meters=meters,
        site_title=title, server_version=version,
        sound_events=sound_events, taps=taps, toggles=toggles)
    return response


def get_version(request):
    return {'server_version': core_util.get_version()}


def get_keg(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    return protolib.ToProto(keg, full=True)


def get_keg_drinks(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    return keg.drinks.all()


def get_keg_events(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    events = keg.events.all()
    events = apply_since(request, events)
    return events


def get_keg_sizes(request):
    # Deprecated endpoint.
    ret = []
    fake_id = 0
    for size_name, volume_ml in keg_sizes.VOLUMES_ML.iteritems():
        ret.append({
            'volume_ml': volume_ml,
            'id': fake_id,
            'description': keg_sizes.DESCRIPTIONS[size_name],
        })
        fake_id += 1
    return ret


@require_http_methods(["POST"])
@auth_required
@csrf_exempt
def end_keg(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    tap = keg.current_tap
    keg = request.backend.end_keg(tap)
    return protolib.ToProto(keg, full=True)


def all_sessions(request):
    return models.DrinkingSession.objects.all()


def current_session(request):
    try:
        latest = models.DrinkingSession.objects.latest()
        if latest.IsActiveNow():
            return latest
    except models.DrinkingSession.DoesNotExist:
        pass

    raise Http404('There is no active session.')


def all_events(request):
    events = models.SystemEvent.objects.all().order_by('-id')
    events = apply_since(request, events)
    events = events[:10]
    return [protolib.ToProto(e, full=True) for e in events]


def apply_since(request, query):
    """Restricts the query to `since` events, if given."""
    since_str = request.GET.get('since')
    if since_str:
        try:
            since = int(since_str)
            return query.filter(id__gt=since)
        except (ValueError, TypeError):
            pass
    return query


@auth_required
def all_sound_events(request):
    return []  # deprecated


def get_keg_sessions(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    sessions = keg.Sessions()
    return sessions


def get_keg_stats(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)
    return keg.get_stats()


def get_system_stats(request):
    return models.KegbotSite.get().get_stats()


@require_http_methods(['GET', 'POST'])
def all_taps(request):
    if request.method == 'POST':
        util.check_api_key(request)
        return create_tap(request)
    return models.KegTap.objects.all().order_by('name')


def create_tap(request):
    form = forms.TapCreateForm(request.POST)
    if form.is_valid():
        return request.backend.create_tap(name=form.cleaned_data['name'])
    raise kbapi.BadRequestError(_form_errors(form))


@auth_required
def user_list(request):
    return models.User.objects.filter(is_active=True).exclude(username='guest').order_by('username')


def get_user(request, username):
    user = get_object_or_404(models.User, username=username)
    return protolib.ToProto(user, full=True)


def get_user_drinks(request, username):
    user = get_object_or_404(models.User, username=username)
    return user.drinks.all()


def get_user_events(request, username):
    user = get_object_or_404(models.User, username=username)
    return user.events.all()


def get_user_stats(request, username):
    user = get_object_or_404(models.User, username=username)
    return user.get_stats()


@auth_required
def get_auth_token(request, auth_device, token_value):
    tok = request.backend.get_auth_token(auth_device, token_value)
    return tok


@csrf_exempt
@auth_required
def assign_auth_token(request, auth_device, token_value):
    if not request.POST:
        raise kbapi.BadRequestError('POST required.')

    form = forms.AssignTokenForm(request.POST)
    if not form.is_valid():
        errors = _form_errors(form)
        raise kbapi.BadRequestError(errors)

    b = request.backend
    username = form.cleaned_data['username']

    user = b.get_user(username)
    if not user:
        raise kbapi.BadRequestError("User does not exist")

    try:
        tok = b.get_auth_token(auth_device, token_value)
    except backend.exceptions.NoTokenError:
        tok = b.create_auth_token(auth_device, token_value, username=username)

    if tok.user != user:
        if tok.user:
            raise kbapi.BadRequestError("Token is already bound to a user")
        tok.user = user
        tok.save()
    return tok


def all_thermo_sensors(request):
    return models.ThermoSensor.objects.all()


def _get_sensor_or_404(request, sensor_name):
    try:
        sensor = models.ThermoSensor.objects.get(raw_name=sensor_name)
    except models.ThermoSensor.DoesNotExist:
        try:
            sensor = models.ThermoSensor.objects.get(nice_name=sensor_name)
        except models.ThermoSensor.DoesNotExist:
            raise Http404
    return sensor


@csrf_exempt
def get_thermo_sensor(request, sensor_name):
    if request.method == 'POST':
        return _thermo_sensor_post(request, sensor_name)
    else:
        return _thermo_sensor_get(request, sensor_name)


def _thermo_sensor_get(request, sensor_name):
    sensor = _get_sensor_or_404(request, sensor_name)
    logs = sensor.thermolog_set.all()
    if not logs:
        last_temp = None
        last_time = None
    else:
        last_temp = logs[0].temp
        last_time = logs[0].time
    res = {
        'sensor': sensor,
        'last_temp': last_temp,
        'last_time': last_time,
    }
    return res


@auth_required
def _thermo_sensor_post(request, sensor_name):
    form = forms.ThermoPostForm(request.POST)
    if not form.is_valid():
        raise kbapi.BadRequestError, _form_errors(form)
    cd = form.cleaned_data
    sensor, created = models.ThermoSensor.objects.get_or_create(raw_name=sensor_name)
    # TODO(mikey): use form fields to compute `when`
    return request.backend.log_sensor_reading(sensor.raw_name, cd['temp_c'])


def get_thermo_sensor_logs(request, sensor_name):
    sensor = _get_sensor_or_404(request, sensor_name)
    return sensor.thermolog_set.all()[:60 * 2]


def get_api_key(request):
    user = request.user
    api_key = ''
    if user and (user.is_staff or user.is_superuser):
        description = request.GET.get('description', '')
        api_key = models.ApiKey.objects.create(description=description)
        api_key = api_key.key
    return {'api_key': api_key}


@csrf_exempt
def tap_detail(request, meter_name_or_id):
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    if request.method == 'POST':
        util.check_api_key(request)
        return _tap_detail_post(request, tap)
    elif request.method == 'GET':
        return _tap_detail_get(request, tap)
    elif request.method == 'DELETE':
        util.check_api_key(request)
        tap.delete()
        return RESULT_OK

    raise kbapi.BadRequestError('Method not supported')


def _tap_detail_get(request, tap):
    return protolib.ToProto(tap, full=True)


@csrf_exempt
@auth_required
def tap_calibrate(request, meter_name_or_id):
    # TODO(mikey): This would make more semantic sense as PATCH /taps/tap-name/,
    # but Django's support for non-POST verbs is poor (specifically wrt request
    # body/form handling).
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    meter = tap.current_meter()
    if not meter:
        raise kbapi.BadRequestError('Tap does not have a meter!')

    form = forms.CalibrateTapForm(request.POST)
    if form.is_valid():
        meter.ticks_per_ml = 1.0 / form.cleaned_data['ml_per_tick']
        meter.save()
        tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)


@csrf_exempt
@auth_required
def tap_spill(request, meter_name_or_id):
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    if not tap.current_keg:
        raise kbapi.BadRequestError('No keg on tap.')
    form = forms.TapSpillForm(request.POST)
    if form.is_valid():
        tap.current_keg.spilled_ml += form.cleaned_data['volume_ml']
        tap.current_keg.save()
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)


@csrf_exempt
@auth_required
def tap_activate(request, meter_name_or_id):
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    form = ChangeKegForm(request.POST)
    if form.is_valid():
        form.save(tap)
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)


@require_http_methods(["POST"])
@csrf_exempt
@auth_required
def tap_connect_meter(request, meter_name_or_id):
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    form = forms.ConnectMeterForm(request.POST)
    if form.is_valid():
        tap = request.backend.connect_meter(tap, form.cleaned_data['meter'])
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)


@require_http_methods(["POST"])
@csrf_exempt
@auth_required
def tap_disconnect_meter(request, meter_name_or_id):
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    tap = request.backend.connect_meter(tap, None)
    return protolib.ToProto(tap, full=True)


@require_http_methods(["POST"])
@csrf_exempt
@auth_required
def tap_connect_toggle(request, meter_name_or_id):
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    form = forms.ConnectToggleForm(request.POST)
    if form.is_valid():
        tap = request.backend.connect_toggle(tap, form.cleaned_data['toggle'])
    else:
        raise kbapi.BadRequestError, _form_errors(form)
    return protolib.ToProto(tap, full=True)


@require_http_methods(["POST"])
@csrf_exempt
@auth_required
def tap_disconnect_toggle(request, meter_name_or_id):
    tap = get_tap_from_meter_name_or_404(meter_name_or_id)
    tap = request.backend.connect_toggle(tap, None)
    return protolib.ToProto(tap, full=True)


@auth_required
def _tap_detail_post(request, tap):
    form = forms.DrinkPostForm(request.POST)
    if not form.is_valid():
        raise kbapi.BadRequestError, _form_errors(form)
    cd = form.cleaned_data
    if cd.get('pour_time') and cd.get('now'):
        pour_time = datetime.datetime.fromtimestamp(cd.get('pour_time'))
        pour_now = datetime.datetime.fromtimestamp(cd.get('now'))
        pour_time_ago = pour_now - pour_time
        pour_time = timezone.now() - pour_time_ago
    else:
        pour_time = None
    duration = cd.get('duration')
    if duration is None:
        duration = 0
    try:
        drink = request.backend.record_drink(tap,
            ticks=cd['ticks'],
            volume_ml=cd.get('volume_ml'),
            username=cd.get('username'),
            pour_time=pour_time,
            duration=duration,
            shout=cd.get('shout'),
            tick_time_series=cd.get('tick_time_series'),
            photo=request.FILES.get('photo', None))
        return protolib.ToProto(drink, full=True)
    except backend.exceptions.BackendError, e:
        raise kbapi.ServerError(str(e))


@csrf_exempt
@auth_required
def cancel_drink(request):
    if request.method != 'POST':
        raise kbapi.BadRequestError('POST required')
    form = forms.CancelDrinkForm(request.POST)
    if not form.is_valid():
        raise kbapi.BadRequestError, _form_errors(form)
    cd = form.cleaned_data
    try:
        res = request.backend.cancel_drink(drink_id=cd.get('id'), spilled=cd.get('spilled', False))
        return protolib.ToProto(res, full=True)
    except backend.exceptions.BackendError, e:
        raise kbapi.ServerError(str(e))


@require_http_methods(["POST"])
@csrf_exempt
def login(request):
    form = AuthenticationForm(data=request.POST)
    if form.is_valid():
        auth_login(request, form.get_user())
        if request.session.test_cookie_worked():
            request.session.delete_test_cookie()
        return RESULT_OK
    else:
        raise kbapi.PermissionDeniedError('Login failed.')


def logout(request):
    auth_logout(request)
    return RESULT_OK


@csrf_exempt
@auth_required
def register(request):
    if not request.POST:
        raise kbapi.BadRequestError('POST required.')
    form = forms.RegisterForm(request.POST)
    errors = {}
    if not form.is_valid():
        errors = _form_errors(form)
    else:
        username = form.cleaned_data['username']
        email = form.cleaned_data.get('email', None)
        password = form.cleaned_data.get('password', None)
        photo = request.FILES.get('photo', None)
        try:
            user = request.backend.create_new_user(username, email=email,
                password=password, photo=photo)
            return protolib.ToProto(user, full=True)
        except backend.exceptions.UserExistsError:
            user_errs = errors.get('username', [])
            user_errs.append('Username not available.')
            errors['username'] = user_errs
    raise kbapi.BadRequestError(errors)


@csrf_exempt
@auth_required
def user_photo(request, username):
    user = get_object_or_404(models.User, username=username)
    if request.method == 'POST':
        return post_user_photo(request, user)
    else:
        return get_user_photo(request, user)


def get_user_photo(request, user):
    mugshot = user.mugshot
    if not mugshot:
        return {}
    return mugshot


def post_user_photo(request, user):
    photo_file = request.FILES.get('photo')
    if not photo_file:
        raise kbapi.BadRequestError('The file "photo" is required.')

    pic = models.Picture.objects.create(user=user)
    pic.image.save(photo_file.name, photo_file)
    pic.save()

    user.mugshot = pic
    user.save()
    return protolib.ToProto(pic)


@csrf_exempt
@require_http_methods(["POST"])
def link_device_new(request):
    name = request.POST.get('name', 'Unknown Device')
    code = devicelink.start_link(name)
    return {'status': 'ok', 'code': code, 'linked': False}


@require_http_methods(["GET"])
def link_device_status(request, code):
    try:
        api_key = devicelink.get_status(code)
    except devicelink.LinkExpiredException:
        raise Http404('Code expired or does not exist.')
    if api_key:
        return {'status': 'ok', 'linked': True, 'api_key': api_key}
    return {'status': 'ok', 'linked': False}


def default_handler(request):
    raise Http404('Not an API endpoint: %s' % request.path[:100])


def get_tap_from_meter_name_or_404(meter_name_or_id):
    try:
        meter_id = int(meter_name_or_id)
        try:
            return models.KegTap.objects.get(pk=meter_id)
        except models.KegTap.DoesNotExist as e:
            raise Http404(str(e))
    except ValueError:
        pass

    try:
        return models.KegTap.get_from_meter_name(meter_name_or_id)
    except models.KegTap.DoesNotExist, e:
        raise Http404(str(e))



from django.conf import settings

import urllib

from pykeg.core import models
from pykeg.core import util
from pykeg.web.kegweb.forms import LoginForm


def kbsite(request):
    kbsite = getattr(request, 'kbsite', None)

    redir = urllib.urlencode({'redir': request.build_absolute_uri(request.path)})

    sso_login_url = getattr(settings, 'SSO_LOGIN_URL', '')
    if sso_login_url:
        sso_login_url = '{}?{}'.format(sso_login_url, redir)

    sso_logout_url = getattr(settings, 'SSO_LOGOUT_URL', '')
    if sso_logout_url:
        sso_logout_url = '{}?{}'.format(sso_logout_url, redir)

    ret = {
        'DEBUG': settings.DEBUG,
        'DEMO_MODE': settings.DEMO_MODE,
        'EMBEDDED': settings.EMBEDDED,
        'VERSION': util.get_version(),
        'HAVE_SESSIONS': False,
        'HAVE_ADMIN': settings.KEGBOT_ENABLE_ADMIN,
        'GOOGLE_ANALYTICS_ID': None,
        'SSO_LOGIN_URL': sso_login_url,
        'SSO_LOGOUT_URL': sso_logout_url,
        'CAN_INVITE': kbsite.can_invite(request.user) if kbsite else False,
        'kbsite': kbsite,
        'request_path': request.path,
        'login_form': LoginForm(initial={'next_page': request.path}),
        'guest_info': {
            'name': 'guest',
            'image': None,
        },
        'PLUGINS': getattr(request, 'plugins', {}),
    }

    if kbsite:
        ret['HAVE_SESSIONS'] = models.DrinkingSession.objects.all().count() > 0
        ret['GOOGLE_ANALYTICS_ID'] = kbsite.google_analytics_id
        ret['metric_volumes'] = (kbsite.volume_display_units == 'metric')
        ret['temperature_display_units'] = kbsite.temperature_display_units

    return ret



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin interface for extending the Kegbot frontend."""

import logging
from pykeg.plugin.datastore import ModelDatastore


class Plugin:
    """Interface class for plugins."""

    NAME = None
    SHORT_NAME = None
    DESCRIPTION = None
    VERSION = None
    URL = None

    def __init__(self, datastore=None, plugin_registry=None):
        self.datastore = datastore if datastore else ModelDatastore(self.get_short_name())
        self.logger = logging.getLogger(self.get_short_name())
        self.plugin_registry = plugin_registry or {}

    @classmethod
    def get_version(cls):
        """Return the curent plugin version.

        Subclasses should set the class attribute VERSION, which is
        returned by the default implementation.

        Returns:
            The version, as a tuple of (major, minor, micro[, tag]).
            Major, minor, and micro must be integers; tag, if present,
            must be an arbitrary string.
        """
        if not cls.VERSION:
            raise NotImplementedError
        return cls.VERSION

    @classmethod
    def get_name(cls):
        """Return a human-readable (one or two word) name for this plugin.

        Subclasses should set the class attribute NAME, which is
        returned by the default implementation.
        """
        if not cls.NAME:
            raise NotImplementedError
        return cls.NAME

    @classmethod
    def get_short_name(cls):
        """Return a short for this plugin.

        Subclasses should set the class attribute NAME, which is
        returned by the default implementation.
        """
        if not cls.SHORT_NAME:
            raise NotImplementedError
        return cls.SHORT_NAME

    @classmethod
    def get_description(cls):
        """Return a concise human-readable description for this plugin.

        Subclasses should set the class attribute DESCRIPTION, which is
        returned by the default implementation.
        """
        if not cls.DESCRIPTION:
            raise NotImplementedError
        return cls.DESCRIPTION

    @classmethod
    def get_url(cls):
        """Returns the information URL (homepage) for the plugin.

        Subclassses should set the class attribute URL, which is returned
        by the default implementation.
        """
        if not cls.URL:
            raise NotImplementedError
        return cls.URL

    ### Plugin methods

    def get_admin_settings_view(self):
        """Returns the view instance for the main admin settings for this
        plugin, or None.
        """
        return None

    def get_extra_admin_views(self):
        """Returns an iterable of additional views to be installed in the admin
        site section this plugin.

        Each item should be a 3-tuple of the form:
          (regex, view name, url name)

        Each view will be installed with the name
        "plugin-<plugin name>-<url name>"
        """
        return []

    def get_user_settings_view(self):
        """Returns the view instance for the main user settings for this
        plugin, or None.
        """
        return None

    def get_extra_user_views(self):
        """Returns an iterable of additional views to be installed in the
        user section for this plugin.

        Each item should be a 3-tuple of the form:
          (regex, view name, url name)

        Each view will be installed with the name
        "plugin-<plugin name>-<url name>"
        """
        return []

    def handle_new_events(self, event):
        """Called synchronously when new events are posted.

        Plugins should *quickly* perform any work. Long-running work can be
        performed by scheduling a background task in this method.
        """
        pass

    ### Helpers

    def save_form(self, form, prefix):
        return self.datastore.save_form(form, prefix)

    def load_form(self, form_cls, prefix, **form_kwargs):
        return self.datastore.load_form(form_cls, prefix, **form_kwargs)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import urlparse

from django.conf import settings
from django.contrib import messages
from django.shortcuts import redirect


class DemoModeMiddleware:
    """Denies non-GET requests when in demo mode.."""
    WHITELISTED_PATHS = (
        '/accounts/login/',
        '/accounts/logout/',
        '/demo/',
    )

    def process_request(self, request):
        if not getattr(settings, 'DEMO_MODE', False):
            return
        if request.method == 'GET':
            return
        for path in self.WHITELISTED_PATHS:
            if request.path.startswith(path):
                return

        messages.error(request, 'Site is in demo mode; changes were not saved.')
        path_or_url = urlparse.urlparse(request.META.get('HTTP_REFERER', '')).path
        if not path_or_url:
            path_or_url = 'kb-home'

        return redirect(path_or_url)



import random
from django.db.models import Count


def random_item(model_cls):
    count = model_cls.objects.aggregate(count=Count('id'))['count']
    if count == 0:
        return None
    random_index = random.randint(0, count - 1)
    return model_cls.objects.all()[random_index]



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Test for util module."""

from pykeg.core import util

from django.test import TestCase


class CoreTests(TestCase):

    def test_get_version(self):
        self.assertNotEqual('0.0.0', util.get_version())
        try:
            util.get_version_object()
        except ValueError as e:
            self.fail('Illegal version: ' + str(e))
        self.assertTrue(util.get_version_object().version >= (0, 9, 23))



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from pykeg.backend import get_kegbot_backend
from pykeg.core import models
from pykeg.core.util import get_version_object
from pykeg.core.util import set_current_request
from pykeg.web.api.util import is_api_request

from pykeg.plugin import util as plugin_util

from django.conf import settings
from django.http import HttpResponse
from django.http import HttpResponseServerError
from django.template.response import SimpleTemplateResponse
from django.template import RequestContext
from django.utils import timezone

import logging

logger = logging.getLogger(__name__)

# Requests are always allowed for these path prefixes.
PRIVACY_EXEMPT_PATHS = (
    '/account/activate/',
    '/accounts/register',
    '/accounts/login/',
    '/admin/',
    '/media/',
    '/setup/',
    '/sso/login',
    '/sso/logout',
)

PRIVACY_EXEMPT_PATHS += getattr(settings, 'KEGBOT_EXTRA_PRIVACY_EXEMPT_PATHS', ())


def _path_allowed(path, kbsite):
    for p in PRIVACY_EXEMPT_PATHS:
        if path.startswith(p):
            return True
    return False


class CurrentRequestMiddleware:
    """Set/clear the current request."""
    def process_request(self, request):
        set_current_request(request)

    def process_response(self, request, response):
        request.kb = None
        set_current_request(None)
        return response


class KegbotSiteMiddleware:
    ALLOWED_VIEW_MODULE_PREFIXES = (
        'pykeg.web.setup_wizard.',
    )

    def process_request(self, request):
        request.need_setup = False
        request.need_upgrade = False
        request.kbsite = None

        installed_version = models.KegbotSite.get_installed_version()
        if installed_version is None:
            request.need_setup = True
        else:
            request.installed_version_string = str(installed_version)
            request.need_upgrade = installed_version < get_version_object()

        if not request.need_setup and not request.need_upgrade:
            request.kbsite = models.KegbotSite.objects.get(name='default')
            if request.kbsite.is_setup:
                timezone.activate(request.kbsite.timezone)
                request.plugins = dict((p.get_short_name(), p) for p in plugin_util.get_plugins().values())
            else:
                request.need_setup = True

        request.backend = get_kegbot_backend()

        return None

    def process_view(self, request, view_func, view_args, view_kwargs):
        for prefix in self.ALLOWED_VIEW_MODULE_PREFIXES:
            if view_func.__module__.startswith(prefix):
                return None

        if is_api_request(request):
            # API endpoints handle "setup required" differently.
            return None

        if request.need_setup:
            return self._setup_required(request)
        elif request.need_upgrade:
            return self._upgrade_required(request)

        return None

    def _setup_required(self, request):
        if settings.EMBEDDED:
            return HttpResponseServerError('Site is not set up.', content_type='text/plain')
        return SimpleTemplateResponse('setup_wizard/setup_required.html',
            context=RequestContext(request), status=403)

    def _upgrade_required(self, request):
        if settings.EMBEDDED:
            return HttpResponseServerError('Site needs upgrade.', content_type='text/plain')
        context = RequestContext(request)
        context['installed_version'] = getattr(request, 'installed_version_string', None)
        return SimpleTemplateResponse('setup_wizard/upgrade_required.html',
            context=context, status=403)


class PrivacyMiddleware:
    """Enforces site privacy settings.

    Must be installed after ApiRequestMiddleware (in request order) to
    access is_kb_api_request attribute.
    """

    def process_view(self, request, view_func, view_args, view_kwargs):
        if not hasattr(request, 'kbsite'):
            return None
        elif _path_allowed(request.path, request.kbsite):
            return None
        elif request.is_kb_api_request:
            # api.middleware will enforce access requirements.
            return None

        privacy = request.kbsite.privacy

        if privacy == 'public':
            return None
        elif privacy == 'staff':
            if not request.user.is_staff:
                return SimpleTemplateResponse('kegweb/staff_only.html',
                    context=RequestContext(request), status=401)
            return None
        elif privacy == 'members':
            if not request.user.is_authenticated or not request.user.is_active:
                return SimpleTemplateResponse('kegweb/members_only.html',
                    context=RequestContext(request), status=401)
            return None

        return HttpResponse('Server misconfigured, unknown privacy setting:%s' % privacy, status=500)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Checks a central server for updates."""

from django.conf import settings
from django.core.cache import cache
from django.utils import timezone

from pykeg.core import models
from pykeg.core import util
from pykeg.core.tasks import core_checkin_task
from pykeg.core.util import SuppressTaskErrors

import datetime
import logging
import os
import requests

FIELD_REG_ID = 'reg_id'
FIELD_PRODUCT = 'product'
FIELD_VERSION = 'version'

FIELD_INTERVAL_MILLIS = 'interval_millis'
FIELD_UPDATE_AVAILABLE = 'update_available'
FIELD_UPDATE_REQUIRED = 'update_required'
FIELD_UPDATE_TITLE = 'update_title'
FIELD_UPDATE_URL = 'update_url'
FIELD_NEWS = 'news'

PRODUCT = 'kegbot-server'

CHECKIN_URL = os.getenv('CHECKIN_URL', None) or 'https://kegbotcheckin.appspot.com/checkin'
CHECKIN_INTERVAL = datetime.timedelta(hours=12)

KEY_LAST_CHECKIN_TIME = 'checkin:last_checkin_time'
KEY_LAST_CHECKIN_RESPONSE = 'checkin:last_checkin_response'

LOGGER = logging.getLogger('checkin')
logging.getLogger('requests').setLevel(logging.WARNING)


class CheckinError(Exception):
    """Base exception."""


def get_last_checkin():
    """Returns tuple of (time, data), or (None, None) if not available."""
    return cache.get(KEY_LAST_CHECKIN_TIME), cache.get(KEY_LAST_CHECKIN_RESPONSE)


def set_last_checkin(when, response_data=None):
    """Sets last checkin date and time."""
    cache.set(KEY_LAST_CHECKIN_TIME, when, timeout=None)
    cache.set(KEY_LAST_CHECKIN_RESPONSE, response_data, timeout=None)


def schedule_checkin(force=False):
    """Schedules a checkin if needed and allowed.

    Args
        force: if True, ignore last checkin time.

    Returns
        True if a checkin was scheduled.
    """
    if settings.EMBEDDED:
        LOGGER.debug('schedule_checkin: not scheduling: site is embedded')
        return False

    kbsite = models.KegbotSite.get()
    if not kbsite.check_for_updates:
        LOGGER.debug('schedule_checkin: not scheduling: checkin disabled')
        return False

    last_checkin_time, last_checkin_data = get_last_checkin()
    now = timezone.now()
    if not last_checkin_time or (now - last_checkin_time) > CHECKIN_INTERVAL or force:
        with SuppressTaskErrors(LOGGER):
            LOGGER.info('schedule_checkin: scheduling checkin')
            core_checkin_task.delay()
            return True

    return False


def checkin(url=CHECKIN_URL, product=PRODUCT, timeout=None, quiet=False):
    """Issue a single checkin to the checkin server.

    No-op if kbsite.check_for_updates is False.

    Returns
        A checkin response dictionary, or None if checkin is disabled.

    Raises
        ValueError: On malformed reponse.
        requests.RequestException: On error talking to server.
    """
    if settings.EMBEDDED:
        LOGGER.debug('Checkin disabled in embedded mode')
        return

    kbsite = models.KegbotSite.get()
    if not kbsite.check_for_updates:
        LOGGER.debug('Upgrade check is disabled')
        return

    site = models.KegbotSite.get()
    reg_id = site.registration_id

    headers = {
        'User-Agent': util.get_user_agent(),
    }
    payload = {
        FIELD_PRODUCT: product,
        FIELD_REG_ID: reg_id,
        FIELD_VERSION: util.get_version(),
    }

    try:
        LOGGER.debug('Checking in, url=%s reg_id=%s' % (url, reg_id))
        result = requests.post(url, data=payload, headers=headers, timeout=timeout).json()
        new_reg_id = result.get(FIELD_REG_ID)
        if new_reg_id != reg_id:
            LOGGER.debug('Updating reg_id=%s' % new_reg_id)
            site.registration_id = new_reg_id
            site.save()
        LOGGER.debug('Checkin result: %s' % str(result))
        if not quiet:
            LOGGER.info('Checkin complete, reg_id=%s' % (reg_id,))
        set_last_checkin(timezone.now(), result)
        return result
    except (ValueError, requests.RequestException) as e:
        if not quiet:
            LOGGER.warning('Checkin error: %s' % str(e))
        raise CheckinError(e)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for pykeg.core.models"""

import datetime
import os

from django.test import TestCase
from django.core.files import File

from . import kb_common
from . import models
from .testutils import make_datetime

from pykeg.backend import get_kegbot_backend
from pykeg.core.testutils import get_filename
from kegbot.util import units


class CoreModelsTestCase(TestCase):
    def setUp(self):
        models.KegbotSite.get()  # create the site
        self.backend = get_kegbot_backend()
        self.producer = models.BeverageProducer.objects.create(
            name='Moonshine Beers',
            country='USA',
            origin_state='Anystate',
            origin_city='Bathtub',
            url='http://example.com/',
            description='Pretty bad beers.',
        )

        self.beverage = models.Beverage.objects.create(
            name='Moonshine Porter',
            producer=self.producer,
            style='Porter',
            calories_per_ml=3.0,
            carbs_per_ml=10.0,
            abv_percent=0.05,
        )

        self.keg = models.Keg.objects.create(
            type=self.beverage,
            keg_type='other',
            start_time=make_datetime(2000, 4, 1),
            end_time=make_datetime(2000, 5, 1),
            description='Our first keg!',
            full_volume_ml=2000,
        )

        self.tap = models.KegTap.objects.create(
            name='Test Tap',
            current_keg=self.keg,
        )

        self.controller = models.Controller.objects.create(
            name='kegboard'
        )

        self.meter = models.FlowMeter.objects.create(
            controller=self.controller,
            port_name='flow0',
            ticks_per_ml=2.2,
            tap=self.tap
        )

        self.user = models.User.objects.create(
            username='kb_tester',
        )

        self.user2 = models.User.objects.create(
            username='kb_tester2',
        )

    def testKegStuff(self):
        """Test basic keg relations that should always work."""
        self.assertEqual(self.keg.full_volume_ml,
            units.Quantity(2.0, units.UNITS.Liter).InMilliliters())
        self.assertEqual(self.keg.type.producer.name, "Moonshine Beers")

        self.assertEqual(0.0, self.keg.served_volume())
        self.assertEqual(2000, self.keg.remaining_volume_ml())

    def testDrinkAccounting(self):
        d = self.backend.record_drink(self.tap,
            ticks=1200,
            username=self.user.username,
                                      )
        self.assertEqual(d.keg.served_volume(), d.volume_ml)

    def testDrinkSessions(self):
        """ Checks for the DrinkingSession records. """
        u1 = self.user
        u2 = self.user2
        units.Quantity(1200)

        base_time = make_datetime(2009, 1, 1, 1, 0, 0)

        td_10m = datetime.timedelta(minutes=10)
        td_400m = datetime.timedelta(minutes=400)
        td_390m = td_400m - td_10m

        self.assertEqual(models.Drink.objects.all().count(), 0)
        self.assertEqual(models.DrinkingSession.objects.all().count(), 0)

        # u=1 t=0
        self.backend.record_drink(self.tap,
            ticks=1200,
            username=u1.username,
            pour_time=base_time,
                                  )
        # u=2 t=0
        self.backend.record_drink(self.tap,
            ticks=1200,
            username=u2.username,
            pour_time=base_time,
                                  )

        # u=1 t=10
        self.backend.record_drink(self.tap,
            ticks=1200,
            username=u1.username,
            pour_time=base_time + td_10m,
                                  )

        # u=1 t=400
        self.backend.record_drink(self.tap,
            ticks=1200,
            username=u1.username,
            pour_time=base_time + td_400m,
                                  )

        # u=2 t=490
        self.backend.record_drink(self.tap,
            ticks=1200,
            username=u2.username,
            pour_time=base_time + td_390m,
                                  )

        # u=2 t=400
        self.backend.record_drink(self.tap,
            ticks=1200,
            username=u2.username,
            pour_time=base_time + td_400m,
                                  )

        drinks_u1 = u1.drinks.all().order_by('time')

        s1, s2 = models.DrinkingSession.objects.all().order_by('start_time')[:2]

        SESSION_DELTA = datetime.timedelta(minutes=kb_common.DRINK_SESSION_TIME_MINUTES)

        # session 1: should be 10 minutes long as created above
        self.assertEqual(s1.start_time, drinks_u1[0].time)
        self.assertEqual(s1.end_time, drinks_u1[0].time + td_10m + SESSION_DELTA)
        self.assertEqual(s1.drinks.all().filter(user=u1).count(), 2)
        self.assertEqual(s1.drinks.all().filter(user=u2).count(), 1)

        # session 2: at time 200, 1 drink
        self.assertEqual(s2.start_time, base_time + td_390m)
        self.assertEqual(s2.end_time, base_time + td_400m + SESSION_DELTA)
        self.assertEqual(s2.drinks.all().filter(user=u1).count(), 1)
        self.assertEqual(s2.drinks.all().filter(user=u2).count(), 2)

        # Now check DrinkingSessions were created correctly; there should be
        # two groups capturing all 4 sessions.
        all_groups = models.DrinkingSession.objects.all().order_by('start_time')
        self.assertEqual(len(all_groups), 2)

        self.assertEqual(all_groups[0].start_time, base_time)
        self.assertEqual(all_groups[0].end_time, base_time + td_10m + SESSION_DELTA)

        self.assertEqual(all_groups[1].start_time, base_time + td_390m)
        self.assertEqual(all_groups[1].end_time, base_time + td_400m + SESSION_DELTA)

        self.assertEqual("kb_tester2 and kb_tester", s2.summarize_drinkers())

    def test_pic_filename(self):
        basename = '1/2/3-4567 89.jpg'
        now = datetime.datetime(2011, 02, 03)
        uuid_str = 'abcdef'
        uploaded_name = models._pics_file_name(None, basename, now, uuid_str)
        self.assertEqual('pics/20110203000000-abcdef.jpg', uploaded_name)

    def test_site_can_invite(self):
        site = models.KegbotSite.get()
        self.assertFalse(site.can_invite(None))

        site.registration_mode = 'public'
        site.save()
        self.assertTrue(site.can_invite(self.user))

        site.registration_mode = 'member-invite-only'
        site.save()
        self.assertTrue(site.can_invite(self.user))

        site.registration_mode = 'staff-invite-only'
        site.save()
        self.assertFalse(site.can_invite(self.user))

        self.user.is_staff = True
        self.user.save()
        site.registration_mode = 'staff-invite-only'
        site.save()
        self.assertTrue(site.can_invite(self.user))

    def test_photos(self):
        picture_file = File(open(get_filename('test_image_800x600.png'), 'rb'))
        picture_obj = models.Picture.objects.create(image=picture_file)

        paths = []
        for spec in ('resized', 'resized_png', 'thumbnail', 'thumbnail_png', 'image'):
            img = getattr(picture_obj, spec)
            path = img.path
            self.assertTrue(os.path.exists(path))
            paths.append(path)

        picture_obj.erase_and_delete()
        for path in paths:
            self.assertFalse(os.path.exists(path))



"""Redis logging utilities.

Adopted from python-redis-log under the following license:

    Copyright (c) 2011 Jed Parsons <jedp@jedparsons.com>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.

Local changes:
    - Added get_current_request() in _getCallingContext()
    - Added request_info in RedisLogRecord
"""

import socket
import getpass
import datetime
import inspect
import logging

from pykeg.core.util import get_current_request


def levelAsString(level):
    return {logging.DEBUG: 'debug',
            logging.INFO: 'info',
            logging.WARNING: 'warning',
            logging.ERROR: 'error',
            logging.CRITICAL: 'critical',
            logging.FATAL: 'fatal'}.get(level, 'unknown')


def _getCallingContext():
    """
    Utility function for the RedisLogRecord.

    Returns the module, function, and lineno of the function
    that called the logger.

    We look way up in the stack.  The stack at this point is:
    [0] logger.py _getCallingContext (hey, that's me!)
    [1] logger.py __init__
    [2] logger.py makeRecord
    [3] _log
    [4] <logging method>
    [5] caller of logging method
    """
    frames = inspect.stack()

    if len(frames) > 4:
        context = frames[5]
    else:
        context = frames[0]

    modname = context[1]
    lineno = context[2]

    if context[3]:
        funcname = context[3]
    else:
        funcname = ""

    request = get_current_request()

    # python docs say you don't want references to
    # frames lying around.  Bad things can happen.
    del context
    del frames

    return modname, funcname, lineno, request


class RedisLogRecord(logging.LogRecord):
    def __init__(self, name, lvl, fn, lno, msg, args, exc_info, func=None, extra=None):
        logging.LogRecord.__init__(self, name, lvl, fn, lno, msg, args, exc_info, func)

        # You can also access the following instance variables via the
        # formatter as
        #     %(hostname)s
        #     %(username)s
        #     %(modname)s
        # etc.
        self.hostname = socket.gethostname()
        self.username = getpass.getuser()
        self.modname, self.funcname, self.lineno, request = _getCallingContext()

        self._raw = {
            'name': name,
            'level': levelAsString(lvl),
            'filename': fn,
            'line_no': self.lineno,
            'msg': unicode(msg),
            'args': list(args),
            'time': datetime.datetime.utcnow(),
            'username': self.username,
            'funcname': self.funcname,
            'hostname': self.hostname,
            'traceback': exc_info,
            'request_info': self._request_info(request),
        }

    def _request_info(self, request):
        """Extracts loggable information from a request."""
        ret = {}
        if not request or not hasattr(request, 'META'):
            return ret
        ret['addr'] = request.META.get('REMOTE_ADDR')
        ret['request_path'] = request.path
        ret['method'] = request.method
        return ret


class RedisLogger(logging.getLoggerClass()):
    def makeRecord(self, name, lvl, fn, lno, msg, args, exc_info, func=None, extra=None):
        record = RedisLogRecord(name, lvl, fn, lno, msg, args, exc_info, func=None)

        if extra:
            for key in extra:
                if (key in ["message", "asctime"]) or (key in record.__dict__):
                    raise KeyError("Attempt to overwrite %r in RedisLogRecord" % key)
                record.__dict__[key] = extra[key]
        return record



"""Redis logging handlers.

Adopted from python-redis-log under the following license:

    Copyright (c) 2011 Jed Parsons <jedp@jedparsons.com>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.

Local changes:
    - Added redis parameters to RedisListHandler
"""

import logging
import redis
from kegbot.util import kbjson as json


class RedisFormatter(logging.Formatter):
    def format(self, record):
        """
        JSON-encode a record for serializing through redis.

        Convert date to iso format, and stringify any exceptions.
        """
        data = record._raw.copy()

        # serialize the datetime date as utc string
        data['time'] = data['time'].isoformat()

        # stringify exception data
        if data.get('traceback'):
            data['traceback'] = self.formatException(data['traceback'])

        return json.dumps(data)


class RedisHandler(logging.Handler):
    """
    Publish messages to redis channel.

    As a convenience, the classmethod to() can be used as a
    constructor, just as in Andrei Savu's mongodb-log handler.
    """

    @classmethod
    def to(cklass, channel, host='localhost', port=6379, password=None, level=logging.NOTSET):
        return cklass(channel, redis.Redis(host=host, port=port, password=password), level=level)

    def __init__(self, channel, redis_client, level=logging.NOTSET):
        """
        Create a new logger for the given channel and redis_client.
        """
        logging.Handler.__init__(self, level)
        self.channel = channel
        self.redis_client = redis_client
        self.formatter = RedisFormatter()

    def emit(self, record):
        """
        Publish record to redis logging channel
        """
        try:
            self.redis_client.publish(self.channel, self.format(record))
        except redis.RedisError:
            pass


class RedisListHandler(logging.Handler):
    """
    Publish messages to redis a redis list.

    As a convenience, the classmethod to() can be used as a
    constructor, just as in Andrei Savu's mongodb-log handler.

    If max_messages is set, trim the list to this many items.
    """

    @classmethod
    def to(cklass, key, max_messages=None, host='localhost', port=6379, level=logging.NOTSET):
        return cklass(key, max_messages, redis.Redis(host=host, port=port), level=level)

    def __init__(self, key, max_messages, redis_client=None,
            redis_host='localhost', redis_port=6379, redis_db=0,
            level=logging.NOTSET):
        """
        Create a new logger for the given key and redis_client.
        """
        logging.Handler.__init__(self, level)
        self.key = key
        if redis_client:
            self.redis_client = redis_client
        else:
            self.redis_client = redis.StrictRedis(host=redis_host, port=redis_port, db=redis_db)
        self.formatter = RedisFormatter()
        self.max_messages = max_messages

    def emit(self, record):
        """
        Publish record to redis logging list
        """
        try:
            if self.max_messages:
                p = self.redis_client.pipeline()
                p.rpush(self.key, self.format(record))
                p.ltrim(self.key, -self.max_messages, -1)
                p.execute()
            else:
                self.redis_client.rpush(self.key, self.format(record))
        except redis.RedisError:
            pass

    def get_logs(self):
        for e in self.redis_client.lrange(self.key, 0, -1):
            try:
                yield json.loads(e)
            except ValueError:
                continue



#!/usr/bin/env python

"""Kegbot Server package.

Kegbot is a hardware and software system to record and monitor access to a beer
kegerator.  For more information and documentation, see http://kegbot.org/
"""

from setuptools import setup, find_packages

VERSION = '1.0.0a2'
DOCLINES = __doc__.split('\n')

SHORT_DESCRIPTION = DOCLINES[0]
LONG_DESCRIPTION = '\n'.join(DOCLINES[2:])
DEPENDENCIES = [
  'kegbot-pyutils == 0.1.7',
  'kegbot-api == 0.1.16',

  'Django == 1.6.5',
  'django-imagekit == 3.1',
  'django-registration == 1.0',
  'django-socialregistration == 0.5.10',
  'django-bootstrap-pagination == 0.1.10',

  'Celery == 3.1.11',
  'kombu == 3.0.16',

  'South == 0.8.4',
  'django-crispy-forms == 1.2.8',
  'foursquare == 2014.04.10',
  'gunicorn == 18.0',
  'MySQL-python == 1.2.5',
  'pillow == 2.4.0',
  'protobuf == 2.5.0',
  'python-gflags == 2.0',
  'django-redis == 3.6.1',
  'pytz == 2014.2',
  'redis == 2.9.1',
  'requests == 2.2.1',
  'tweepy == 2.2',
  'jsonfield == 0.9.20',
]

def setup_package():
  setup(
      name = 'kegbot',
      version = VERSION,
      description = SHORT_DESCRIPTION,
      long_description = LONG_DESCRIPTION,
      author = 'Bevbot LLC',
      author_email = 'info@bevbot.com',
      url = 'http://kegbot.org/',
      packages = find_packages(),
      scripts = [
        'bin/kegbot',
        'bin/setup-kegbot.py',
      ],
      install_requires = DEPENDENCIES,
      dependency_links = [
          'https://github.com/rem/python-protobuf/tarball/master#egg=protobuf-2.4.1',
      ],
      include_package_data = True,
      entry_points = {
        'console_scripts': ['instance=django.core.management:execute_manager'],
      },
  )

if __name__ == '__main__':
  setup_package()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Methods to generate cached statistics from drinks."""

import copy
import inspect
import logging

from pykeg.core import models
from kegbot.util import util

STAT_MAP = {}

logger = logging.getLogger(__name__)


class StatsBuilder:
    """Derives statistics from drinks."""

    def __init__(self):
        self.functions = []
        for name, fn in inspect.getmembers(self, inspect.ismethod):
            if not name.startswith('_') and name != 'build':
                self.functions.append((name, fn))

    def build(self, drink, previous_stats):
        if previous_stats is None:
            previous_stats = util.AttrDict()

        logger.debug('build: drink={}'.format(drink.id))
        stats = util.AttrDict()

        for statname, fn in self.functions:
            previous_value = previous_stats.get(statname, None)
            if previous_value is not None:
                val = fn(drink, previous_stats, previous_value=previous_value)
            else:
                val = fn(drink, previous_stats)

            if val is None:
                raise ValueError('Stat generator for %s returned None' % statname)
            stats[statname] = val
        return stats

    def last_drink_id(self, drink, previous_stats, previous_value=0):
        return drink.id

    def keg_ids(self, drink, previous_stats, previous_value=[]):
        ret = copy.copy(previous_value)
        if drink.keg.id not in ret:
            ret.append(drink.keg.id)
        return ret

    def total_volume_ml(self, drink, previous_stats, previous_value=0):
        return previous_value + drink.volume_ml

    def total_pours(self, drink, previous_stats, previous_value=0):
        return previous_value + 1

    def average_volume_ml(self, drink, previous_stats, previous_value=0):
        vol = previous_stats.get('total_volume_ml', 0)
        count = previous_stats.get('total_pours', 0)
        vol += drink.volume_ml
        count += 1
        return vol / float(count)

    def greatest_volume_ml(self, drink, previous_stats, previous_value=0):
        return float(max(drink.volume_ml, previous_value))

    def greatest_volume_id(self, drink, previous_stats, previous_value=0):
        if drink.volume_ml > previous_stats.get('greatest_volume_ml', 0):
            return drink.id
        return previous_value

    def volume_by_day_of_week(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        drink_weekday = str(drink.session.start_time.strftime('%w'))
        ret[drink_weekday] = ret.get(drink_weekday, 0) + drink.volume_ml
        return ret

    def registered_drinkers(self, drink, previous_stats, previous_value=[]):
        ret = copy.copy(previous_value)
        if drink.user:
            user_id = str(drink.user.id)
            if user_id not in ret:
                ret.append(user_id)
        return ret

    def sessions_count(self, drink, previous_stats, previous_value=0):
        # Use volume_by_session, ensuring our session is captured
        # by injecting a dummy value.
        prev_sessions = previous_stats.get('volume_by_session', {})
        ret = len(prev_sessions.keys())
        if str(drink.session.id) not in prev_sessions:
            ret += 1
        return ret

    def volume_by_year(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        year = str(drink.time.year)
        ret[year] = ret.get(year, 0) + drink.volume_ml
        return ret

    def has_guest_pour(self, drink, previous_stats, previous_value=False):
        if previous_value:
            return True
        return drink.is_guest_pour()

    def volume_by_drinker(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        user_id = str(drink.user.id)
        orig = ret.get(user_id, 0)
        ret[user_id] = float(orig + drink.volume_ml)
        return ret

    def volume_by_session(self, drink, previous_stats, previous_value={}):
        ret = copy.copy(previous_value)
        session_id = str(drink.session.id)
        ret[session_id] = ret.get(session_id, 0) + drink.volume_ml
        return ret

    def largest_session(self, drink, previous_stats, previous_value={}):
        if drink.session.volume_ml >= previous_value.get('volume_ml', 0):
            return {
                'session_id': drink.session.id,
                'volume_ml': drink.session.volume_ml,
            }
        return previous_value

BUILDER = StatsBuilder()

### Public methods


def invalidate(drink_id):
    """Clears all statistics starting from (and including) drink_id."""
    logger.debug('--- Invalidating stats since id {}'.format(drink_id))
    models.Stats.objects.filter(drink_id__gte=drink_id).delete()


def invalidate_all():
    logger.debug('--- Invalidating ALL stats')
    models.Stats.objects.all().delete()


def build_for_id(drink_id):
    """Build stats up to and including this drink id, *without*
    considering any subsequent drinks.
    """
    try:
        drink = models.Drink.objects.get(pk=drink_id)
    except models.Drink.DoesNotExist:
        logger.warning('No drink exists with id={}'.format(drink_id))
        return
    generate(drink, invalidate_first=False)


def rebuild_from_id(drink_id):
    """Builds statistics stating from `drink_id`.  Unlike `build_for_id()`,
    this method computes statistics for any subsequent drinks found in the
    database.
    """
    invalidate(drink_id)
    for drink in models.Drink.objects.filter(id__gte=drink_id).order_by('id'):
        generate(drink, invalidate_first=False)


def _generate_view(drink, user, session, keg):
    """Generates a single "view" (row) based on `drink`."""
    logger.debug('>>> Building stats for drink={}: user={} session={} keg={}'.format(
        drink.id,
        user.username if user else None,
        session.id if session else None,
        keg.id if keg else None))

    # Determine previous drinks in the same view.  This *may* be empty if this
    # is the first drink for the view (first pour on the system, drink, etc.)
    prior_drinks_in_view = models.Drink.objects.filter(id__lt=drink.id)
    if user:
        prior_drinks_in_view = prior_drinks_in_view.filter(user=user.id)
    if session:
        prior_drinks_in_view = prior_drinks_in_view.filter(session=session.id)
    if keg:
        prior_drinks_in_view = prior_drinks_in_view.filter(keg=keg.id)

    prior_drinks_in_view = prior_drinks_in_view.order_by('-id')

    prior_stats = util.AttrDict()
    build_list = [drink]
    if prior_drinks_in_view.count():
        # Starting with the most recent prior drink, get its stats row.
        # If stats don't exist, add drink to build list and continue
        # until a stats row is found or all drinks are exhausted.
        # N.B. we avoid a recursive algorithm due to potentially large recursion
        # depth in certain cases.
        for prior_drink in prior_drinks_in_view:
            try:
                prior_stats = util.AttrDict(
                    models.Stats.objects.get(drink=prior_drink, user=user,
                        session=session, keg=keg).stats
                )
                break
            except models.Stats.DoesNotExist:
                prior_stats = util.AttrDict()
                build_list.insert(0, prior_drink)

    # Build all drinks on the hit list.
    for build_drink in build_list:
        logger.debug('  - operating on drink {}'.format(build_drink.id))
        stats = BUILDER.build(drink=build_drink, previous_stats=prior_stats)
        models.Stats.objects.create(drink=build_drink, user=user, session=session,
            keg=keg, stats=stats, is_first=(not prior_stats))
        prior_stats = stats
    logger.debug('<<< Done.')


def generate(drink, invalidate_first=True):
    """Generates all stats "views" based on `drink`."""
    if invalidate_first:
        invalidate(drink.id)

    # Create an entry for each stats "view".
    for user in (None, drink.user):
        for session in (None, drink.session):
            for keg in (None, drink.keg):
                _generate_view(drink, user, session, keg)

if __name__ == '__main__':
    import cProfile
    command = """main()"""
    cProfile.runctx(command, globals(), locals(), filename="stats.profile")
